# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: record.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0crecord.proto\x12\x07Records\"\\\n\nRecordType\x12\x14\n\x0crecordTypeId\x18\x01 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12\'\n\x05scope\x18\x03 \x01(\x0e\x32\x18.Records.RecordTypeScope\"U\n\x12RecordTypesRequest\x12\x10\n\x08standard\x18\x01 \x01(\x08\x12\x0c\n\x04user\x18\x02 \x01(\x08\x12\x12\n\nenterprise\x18\x03 \x01(\x08\x12\x0b\n\x03pam\x18\x04 \x01(\x08\"\x9c\x01\n\x13RecordTypesResponse\x12(\n\x0brecordTypes\x18\x01 \x03(\x0b\x32\x13.Records.RecordType\x12\x17\n\x0fstandardCounter\x18\x02 \x01(\x05\x12\x13\n\x0buserCounter\x18\x03 \x01(\x05\x12\x19\n\x11\x65nterpriseCounter\x18\x04 \x01(\x05\x12\x12\n\npamCounter\x18\x05 \x01(\x05\"A\n\x18RecordTypeModifyResponse\x12\x14\n\x0crecordTypeId\x18\x01 \x01(\x05\x12\x0f\n\x07\x63ounter\x18\x02 \x01(\x05\"=\n\x11RecordsGetRequest\x12\x13\n\x0brecord_uids\x18\x01 \x03(\x0c\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\xd1\x01\n\x06Record\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12/\n\x0frecord_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\r\n\x05\x65xtra\x18\x05 \x01(\x0c\x12\x0f\n\x07version\x18\x06 \x01(\x05\x12\x1c\n\x14\x63lient_modified_time\x18\x07 \x01(\x03\x12\x10\n\x08revision\x18\x08 \x01(\x03\x12\x10\n\x08\x66ile_ids\x18\t \x03(\x0c\"M\n\x0f\x46olderRecordKey\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_uid\x18\x02 \x01(\x0c\x12\x12\n\nrecord_key\x18\x03 \x01(\x0c\"a\n\x06\x46older\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x12\n\nfolder_key\x18\x02 \x01(\x0c\x12/\n\x0f\x66older_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\x95\x01\n\x04Team\x12\x10\n\x08team_uid\x18\x01 \x01(\x0c\x12\x10\n\x08team_key\x18\x02 \x01(\x0c\x12\x18\n\x10team_private_key\x18\x03 \x01(\x0c\x12-\n\rteam_key_type\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12 \n\x07\x66olders\x18\x05 \x03(\x0b\x32\x0f.Records.Folder\"\xac\x01\n\x12RecordsGetResponse\x12 \n\x07records\x18\x01 \x03(\x0b\x32\x0f.Records.Record\x12\x34\n\x12\x66older_record_keys\x18\x02 \x03(\x0b\x32\x18.Records.FolderRecordKey\x12 \n\x07\x66olders\x18\x03 \x03(\x0b\x32\x0f.Records.Folder\x12\x1c\n\x05teams\x18\x04 \x03(\x0b\x32\r.Records.Team\"4\n\nRecordLink\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\",\n\x0bRecordAudit\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x1c\n\x0cSecurityData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"!\n\x11SecurityScoreData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x84\x03\n\tRecordAdd\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12.\n\x0b\x66older_type\x18\x06 \x01(\x0e\x32\x19.Records.RecordFolderType\x12\x12\n\nfolder_uid\x18\x07 \x01(\x0c\x12\x12\n\nfolder_key\x18\x08 \x01(\x0c\x12)\n\x0crecord_links\x18\t \x03(\x0b\x32\x13.Records.RecordLink\x12#\n\x05\x61udit\x18\n \x01(\x0b\x32\x14.Records.RecordAudit\x12+\n\x0csecurityData\x18\x0b \x01(\x0b\x32\x15.Records.SecurityData\x12\x35\n\x11securityScoreData\x18\x0c \x01(\x0b\x32\x1a.Records.SecurityScoreData\"\x85\x01\n\x11RecordsAddRequest\x12#\n\x07records\x18\x01 \x03(\x0b\x32\x12.Records.RecordAdd\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\x12\x36\n\x16security_data_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\xce\x02\n\x0cRecordUpdate\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x02 \x01(\x03\x12\x10\n\x08revision\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12-\n\x10record_links_add\x18\x06 \x03(\x0b\x32\x13.Records.RecordLink\x12\x1b\n\x13record_links_remove\x18\x07 \x03(\x0c\x12#\n\x05\x61udit\x18\x08 \x01(\x0b\x32\x14.Records.RecordAudit\x12+\n\x0csecurityData\x18\t \x01(\x0b\x32\x15.Records.SecurityData\x12\x35\n\x11securityScoreData\x18\n \x01(\x0b\x32\x1a.Records.SecurityScoreData\"\x8b\x01\n\x14RecordsUpdateRequest\x12&\n\x07records\x18\x01 \x03(\x0b\x32\x15.Records.RecordUpdate\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\x12\x36\n\x16security_data_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\x8e\x01\n\x17RecordFileForConversion\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x14\n\x0c\x66ile_file_id\x18\x02 \x01(\t\x12\x15\n\rthumb_file_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x12\n\nrecord_key\x18\x05 \x01(\x0c\x12\x10\n\x08link_key\x18\x06 \x01(\x0c\"J\n\x19RecordFolderForConversion\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x19\n\x11record_folder_key\x18\x02 \x01(\x0c\"\x92\x02\n\x11RecordConvertToV3\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x02 \x01(\x03\x12\x10\n\x08revision\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12#\n\x05\x61udit\x18\x06 \x01(\x0b\x32\x14.Records.RecordAudit\x12\x35\n\x0brecord_file\x18\x07 \x03(\x0b\x32 .Records.RecordFileForConversion\x12\x36\n\nfolder_key\x18\x08 \x03(\x0b\x32\".Records.RecordFolderForConversion\"]\n\x19RecordsConvertToV3Request\x12+\n\x07records\x18\x01 \x03(\x0b\x32\x1a.Records.RecordConvertToV3\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\'\n\x14RecordsRemoveRequest\x12\x0f\n\x07records\x18\x01 \x03(\x0c\">\n\x0cRecordRevert\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1a\n\x12revert_to_revision\x18\x02 \x01(\x03\">\n\x14RecordsRevertRequest\x12&\n\x07records\x18\x01 \x03(\x0b\x32\x15.Records.RecordRevert\"c\n\x0fRecordLinkError\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12+\n\x06status\x18\x02 \x01(\x0e\x32\x1b.Records.RecordModifyResult\x12\x0f\n\x07message\x18\x03 \x01(\t\"\x95\x01\n\x12RecordModifyStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12+\n\x06status\x18\x02 \x01(\x0e\x32\x1b.Records.RecordModifyResult\x12\x0f\n\x07message\x18\x03 \x01(\t\x12-\n\x0blink_errors\x18\x04 \x03(\x0b\x32\x18.Records.RecordLinkError\"W\n\x15RecordsModifyResponse\x12,\n\x07records\x18\x01 \x03(\x0b\x32\x1b.Records.RecordModifyStatus\x12\x10\n\x08revision\x18\x02 \x01(\x03\"Y\n\x12RecordAddAuditData\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x0f\n\x07version\x18\x04 \x01(\x05\"C\n\x13\x41\x64\x64\x41uditDataRequest\x12,\n\x07records\x18\x01 \x03(\x0b\x32\x1b.Records.RecordAddAuditData\"t\n\x04\x46ile\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x10\n\x08\x66ileSize\x18\x04 \x01(\x03\x12\x11\n\tthumbSize\x18\x05 \x01(\x05\x12\x11\n\tis_script\x18\x06 \x01(\x08\"D\n\x0f\x46ilesAddRequest\x12\x1c\n\x05\x66iles\x18\x01 \x03(\x0b\x32\r.Records.File\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\xa7\x01\n\rFileAddStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12&\n\x06status\x18\x02 \x01(\x0e\x32\x16.Records.FileAddResult\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x12\n\nparameters\x18\x04 \x01(\t\x12\x1c\n\x14thumbnail_parameters\x18\x05 \x01(\t\x12\x1b\n\x13success_status_code\x18\x06 \x01(\x05\"K\n\x10\x46ilesAddResponse\x12%\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x16.Records.FileAddStatus\x12\x10\n\x08revision\x18\x02 \x01(\x03\"f\n\x0f\x46ilesGetRequest\x12\x13\n\x0brecord_uids\x18\x01 \x03(\x0c\x12\x16\n\x0e\x66or_thumbnails\x18\x02 \x01(\x08\x12&\n\x1e\x65mergency_access_account_owner\x18\x03 \x01(\t\"\xa2\x01\n\rFileGetStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12&\n\x06status\x18\x02 \x01(\x0e\x32\x16.Records.FileGetResult\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x1b\n\x13success_status_code\x18\x04 \x01(\x05\x12+\n\x0b\x66ileKeyType\x18\x05 \x01(\x0e\x32\x16.Records.RecordKeyType\"9\n\x10\x46ilesGetResponse\x12%\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x16.Records.FileGetStatus\"\x8d\x01\n\x15\x41pplicationAddRequest\x12\x0f\n\x07\x61pp_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12#\n\x05\x61udit\x18\x05 \x01(\x0b\x32\x14.Records.RecordAudit\"\x88\x01\n\"GetRecordDataWithAccessInfoRequest\x12\x12\n\nclientTime\x18\x01 \x01(\x03\x12\x11\n\trecordUid\x18\x02 \x03(\x0c\x12;\n\x14recordDetailsInclude\x18\x03 \x01(\x0e\x32\x1d.Records.RecordDetailsInclude\"\x86\x02\n\x0eUserPermission\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x12\n\nshareAdmin\x18\x03 \x01(\x08\x12\x10\n\x08sharable\x18\x04 \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x05 \x01(\x08\x12\x18\n\x10\x61waitingApproval\x18\x06 \x01(\x08\x12\x12\n\nexpiration\x18\x07 \x01(\x03\x12\x12\n\naccountUid\x18\x08 \x01(\x0c\x12=\n\x15timerNotificationType\x18\t \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\n \x01(\x08\"\xd8\x01\n\x16SharedFolderPermission\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x12\n\nresharable\x18\x02 \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x03 \x01(\x08\x12\x10\n\x08revision\x18\x04 \x01(\x03\x12\x12\n\nexpiration\x18\x05 \x01(\x03\x12=\n\x15timerNotificationType\x18\x06 \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\x07 \x01(\x08\"\xe8\x02\n\nRecordData\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12\x0f\n\x07version\x18\x02 \x01(\x05\x12\x0e\n\x06shared\x18\x03 \x01(\x08\x12\x1b\n\x13\x65ncryptedRecordData\x18\x04 \x01(\t\x12\x1a\n\x12\x65ncryptedExtraData\x18\x05 \x01(\t\x12\x1a\n\x12\x63lientModifiedTime\x18\x06 \x01(\x03\x12\x15\n\rnonSharedData\x18\x07 \x01(\t\x12-\n\x10linkedRecordData\x18\x08 \x03(\x0b\x32\x13.Records.RecordData\x12\x0e\n\x06\x66ileId\x18\t \x03(\x0c\x12\x10\n\x08\x66ileSize\x18\n \x01(\x03\x12\x15\n\rthumbnailSize\x18\x0b \x01(\x03\x12-\n\rrecordKeyType\x18\x0c \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x11\n\trecordKey\x18\r \x01(\x0c\x12\x11\n\trecordUid\x18\x0e \x01(\x0c\"\xc8\x01\n\x18RecordDataWithAccessInfo\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\'\n\nrecordData\x18\x02 \x01(\x0b\x32\x13.Records.RecordData\x12/\n\x0euserPermission\x18\x03 \x03(\x0b\x32\x17.Records.UserPermission\x12?\n\x16sharedFolderPermission\x18\x04 \x03(\x0b\x32\x1f.Records.SharedFolderPermission\"\x89\x01\n#GetRecordDataWithAccessInfoResponse\x12\x43\n\x18recordDataWithAccessInfo\x18\x01 \x03(\x0b\x32!.Records.RecordDataWithAccessInfo\x12\x1d\n\x15noPermissionRecordUid\x18\x02 \x03(\x0c\"j\n\x12IsObjectShareAdmin\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x0f\n\x07isAdmin\x18\x02 \x01(\x08\x12\x36\n\nobjectType\x18\x03 \x01(\x0e\x32\".Records.CheckShareAdminObjectType\"H\n\rAmIShareAdmin\x12\x37\n\x12isObjectShareAdmin\x18\x01 \x03(\x0b\x32\x1b.Records.IsObjectShareAdmin\"\xbc\x01\n\x18RecordShareUpdateRequest\x12.\n\x0f\x61\x64\x64SharedRecord\x18\x01 \x03(\x0b\x32\x15.Records.SharedRecord\x12\x31\n\x12updateSharedRecord\x18\x02 \x03(\x0b\x32\x15.Records.SharedRecord\x12\x31\n\x12removeSharedRecord\x18\x03 \x03(\x0b\x32\x15.Records.SharedRecord\x12\n\n\x02pt\x18\x04 \x01(\t\"\xc4\x02\n\x0cSharedRecord\x12\x12\n\ntoUsername\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x04 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x05 \x01(\x0c\x12\x10\n\x08\x65\x64itable\x18\x06 \x01(\x08\x12\x11\n\tshareable\x18\x07 \x01(\x08\x12\x10\n\x08transfer\x18\x08 \x01(\x08\x12\x11\n\tuseEccKey\x18\t \x01(\x08\x12\x17\n\x0fremoveVaultData\x18\n \x01(\x08\x12\x12\n\nexpiration\x18\x0b \x01(\x03\x12=\n\x15timerNotificationType\x18\x0c \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\r \x01(\x08\"\xd5\x01\n\x19RecordShareUpdateResponse\x12:\n\x15\x61\x64\x64SharedRecordStatus\x18\x01 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\x12=\n\x18updateSharedRecordStatus\x18\x02 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\x12=\n\x18removeSharedRecordStatus\x18\x03 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\"Z\n\x12SharedRecordStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x10\n\x08username\x18\x04 \x01(\t\"G\n\x1bGetRecordPermissionsRequest\x12\x12\n\nrecordUids\x18\x01 \x03(\x0c\x12\x14\n\x0cisShareAdmin\x18\x02 \x01(\x08\"T\n\x1cGetRecordPermissionsResponse\x12\x34\n\x11recordPermissions\x18\x01 \x03(\x0b\x32\x19.Records.RecordPermission\"l\n\x10RecordPermission\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x03 \x01(\x08\x12\x10\n\x08\x63\x61nShare\x18\x04 \x01(\x08\x12\x13\n\x0b\x63\x61nTransfer\x18\x05 \x01(\x08\"h\n\x16GetShareObjectsRequest\x12\x11\n\tstartWith\x18\x01 \x01(\t\x12\x10\n\x08\x63ontains\x18\x02 \x01(\t\x12\x10\n\x08\x66iltered\x18\x03 \x01(\x08\x12\x17\n\x0fsharedFolderUid\x18\x04 \x01(\x0c\"\xe7\x02\n\x17GetShareObjectsResponse\x12.\n\x12shareRelationships\x18\x01 \x03(\x0b\x32\x12.Records.ShareUser\x12,\n\x10shareFamilyUsers\x18\x02 \x03(\x0b\x32\x12.Records.ShareUser\x12\x30\n\x14shareEnterpriseUsers\x18\x03 \x03(\x0b\x32\x12.Records.ShareUser\x12&\n\nshareTeams\x18\x04 \x03(\x0b\x32\x12.Records.ShareTeam\x12(\n\x0cshareMCTeams\x18\x05 \x03(\x0b\x32\x12.Records.ShareTeam\x12\x32\n\x16shareMCEnterpriseUsers\x18\x06 \x03(\x0b\x32\x12.Records.ShareUser\x12\x36\n\x14shareEnterpriseNames\x18\x07 \x03(\x0b\x32\x18.Records.ShareEnterprise\"\xa5\x01\n\tShareUser\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08\x66ullname\x18\x02 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x03 \x01(\x05\x12$\n\x06status\x18\x04 \x01(\x0e\x32\x14.Records.ShareStatus\x12\x14\n\x0cisShareAdmin\x18\x05 \x01(\x08\x12\"\n\x1aisAdminOfSharedFolderOwner\x18\x06 \x01(\x08\"D\n\tShareTeam\x12\x10\n\x08teamname\x18\x01 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x02 \x01(\x05\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\"?\n\x0fShareEnterprise\x12\x16\n\x0e\x65nterprisename\x18\x01 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x02 \x01(\x05\"S\n\x1fRecordsOnwershipTransferRequest\x12\x30\n\x0ftransferRecords\x18\x01 \x03(\x0b\x32\x17.Records.TransferRecord\"[\n\x0eTransferRecord\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x11\n\tuseEccKey\x18\x04 \x01(\x08\"_\n RecordsOnwershipTransferResponse\x12;\n\x14transferRecordStatus\x18\x01 \x03(\x0b\x32\x1d.Records.TransferRecordStatus\"\\\n\x14TransferRecordStatus\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"y\n\x15RecordsUnshareRequest\x12\x34\n\rsharedFolders\x18\x01 \x03(\x0b\x32\x1d.Records.RecordsUnshareFolder\x12*\n\x05users\x18\x02 \x03(\x0b\x32\x1b.Records.RecordsUnshareUser\"\x86\x01\n\x16RecordsUnshareResponse\x12:\n\rsharedFolders\x18\x01 \x03(\x0b\x32#.Records.RecordsUnshareFolderStatus\x12\x30\n\x05users\x18\x02 \x03(\x0b\x32!.Records.RecordsUnshareUserStatus\"B\n\x14RecordsUnshareFolder\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\";\n\x12RecordsUnshareUser\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x12\n\naccountUid\x18\x02 \x01(\x0c\"H\n\x1aRecordsUnshareFolderStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\"A\n\x18RecordsUnshareUserStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x12\n\naccountUid\x18\x02 \x01(\x0c\"[\n\x1aTimedAccessCallbackPayload\x12=\n\x15timeLimitedAccessType\x18\x01 \x01(\x0e\x32\x1e.Records.TimeLimitedAccessType\"\xfd\x01\n\x18TimeLimitedAccessRequest\x12\x12\n\naccountUid\x18\x01 \x03(\x0c\x12\x0f\n\x07teamUid\x18\x02 \x03(\x0c\x12\x11\n\trecordUid\x18\x03 \x03(\x0c\x12\x17\n\x0fsharedObjectUid\x18\x04 \x01(\x0c\x12=\n\x15timeLimitedAccessType\x18\x05 \x01(\x0e\x32\x1e.Records.TimeLimitedAccessType\x12\x12\n\nexpiration\x18\x06 \x01(\x03\x12=\n\x15timerNotificationType\x18\x07 \x01(\x0e\x32\x1e.Records.TimerNotificationType\"7\n\x17TimeLimitedAccessStatus\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\t\"\xe3\x01\n\x19TimeLimitedAccessResponse\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12:\n\x10userAccessStatus\x18\x02 \x03(\x0b\x32 .Records.TimeLimitedAccessStatus\x12:\n\x10teamAccessStatus\x18\x03 \x03(\x0b\x32 .Records.TimeLimitedAccessStatus\x12<\n\x12recordAccessStatus\x18\x04 \x03(\x0b\x32 .Records.TimeLimitedAccessStatus*h\n\x0fRecordTypeScope\x12\x0f\n\x0bRT_STANDARD\x10\x00\x12\x0b\n\x07RT_USER\x10\x01\x12\x11\n\rRT_ENTERPRISE\x10\x02\x12\n\n\x06RT_PAM\x10\x03\x12\x18\n\x14RT_PAM_CONFIGURATION\x10\x04*\xd1\x01\n\rRecordKeyType\x12\n\n\x06NO_KEY\x10\x00\x12\x19\n\x15\x45NCRYPTED_BY_DATA_KEY\x10\x01\x12\x1b\n\x17\x45NCRYPTED_BY_PUBLIC_KEY\x10\x02\x12\x1d\n\x19\x45NCRYPTED_BY_DATA_KEY_GCM\x10\x03\x12\x1f\n\x1b\x45NCRYPTED_BY_PUBLIC_KEY_ECC\x10\x04\x12\x1d\n\x19\x45NCRYPTED_BY_ROOT_KEY_CBC\x10\x05\x12\x1d\n\x19\x45NCRYPTED_BY_ROOT_KEY_GCM\x10\x06*P\n\x10RecordFolderType\x12\x0f\n\x0buser_folder\x10\x00\x12\x11\n\rshared_folder\x10\x01\x12\x18\n\x14shared_folder_folder\x10\x02*\xec\x02\n\x12RecordModifyResult\x12\x0e\n\nRS_SUCCESS\x10\x00\x12\x12\n\x0eRS_OUT_OF_SYNC\x10\x01\x12\x14\n\x10RS_ACCESS_DENIED\x10\x02\x12\x13\n\x0fRS_SHARE_DENIED\x10\x03\x12\x14\n\x10RS_RECORD_EXISTS\x10\x04\x12\x1e\n\x1aRS_OLD_RECORD_VERSION_TYPE\x10\x05\x12\x1e\n\x1aRS_NEW_RECORD_VERSION_TYPE\x10\x06\x12\x16\n\x12RS_FILES_NOT_MATCH\x10\x07\x12\x1b\n\x17RS_RECORD_NOT_SHAREABLE\x10\x08\x12\x1f\n\x1bRS_ATTACHMENT_NOT_SHAREABLE\x10\t\x12\x19\n\x15RS_FILE_LIMIT_REACHED\x10\n\x12\x1a\n\x16RS_SIZE_EXCEEDED_LIMIT\x10\x0b\x12$\n RS_ONLY_OWNER_CAN_MODIFY_SCRIPTS\x10\x0c*-\n\rFileAddResult\x12\x0e\n\nFA_SUCCESS\x10\x00\x12\x0c\n\x08\x46\x41_ERROR\x10\x01*C\n\rFileGetResult\x12\x0e\n\nFG_SUCCESS\x10\x00\x12\x0c\n\x08\x46G_ERROR\x10\x01\x12\x14\n\x10\x46G_ACCESS_DENIED\x10\x02*J\n\x14RecordDetailsInclude\x12\x13\n\x0f\x44\x41TA_PLUS_SHARE\x10\x00\x12\r\n\tDATA_ONLY\x10\x01\x12\x0e\n\nSHARE_ONLY\x10\x02*b\n\x19\x43heckShareAdminObjectType\x12\x19\n\x15\x43HECK_SA_INVALID_TYPE\x10\x00\x12\x12\n\x0e\x43HECK_SA_ON_SF\x10\x01\x12\x16\n\x12\x43HECK_SA_ON_RECORD\x10\x02*1\n\x0bShareStatus\x12\n\n\x06\x41\x43TIVE\x10\x00\x12\t\n\x05\x42LOCK\x10\x01\x12\x0b\n\x07INVITED\x10\x02*:\n\x15RecordTransactionType\x12\x0f\n\x0bRTT_GENERAL\x10\x00\x12\x10\n\x0cRTT_ROTATION\x10\x01*\xe6\x01\n\x15TimeLimitedAccessType\x12$\n INVALID_TIME_LIMITED_ACCESS_TYPE\x10\x00\x12\x19\n\x15USER_ACCESS_TO_RECORD\x10\x01\x12\'\n#USER_OR_TEAM_ACCESS_TO_SHAREDFOLDER\x10\x02\x12!\n\x1dRECORD_ACCESS_TO_SHAREDFOLDER\x10\x03\x12\x1f\n\x1bUSER_ACCESS_TO_SHAREDFOLDER\x10\x04\x12\x1f\n\x1bTEAM_ACCESS_TO_SHAREDFOLDER\x10\x05*\\\n\x15TimerNotificationType\x12\x14\n\x10NOTIFICATION_OFF\x10\x00\x12\x10\n\x0cNOTIFY_OWNER\x10\x01\x12\x1b\n\x17NOTIFY_PRIVILEGED_USERS\x10\x02\x42#\n\x18\x63om.keepersecurity.protoB\x07Recordsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'record_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\007Records'
  _globals['_RECORDTYPESCOPE']._serialized_start=9466
  _globals['_RECORDTYPESCOPE']._serialized_end=9570
  _globals['_RECORDKEYTYPE']._serialized_start=9573
  _globals['_RECORDKEYTYPE']._serialized_end=9782
  _globals['_RECORDFOLDERTYPE']._serialized_start=9784
  _globals['_RECORDFOLDERTYPE']._serialized_end=9864
  _globals['_RECORDMODIFYRESULT']._serialized_start=9867
  _globals['_RECORDMODIFYRESULT']._serialized_end=10231
  _globals['_FILEADDRESULT']._serialized_start=10233
  _globals['_FILEADDRESULT']._serialized_end=10278
  _globals['_FILEGETRESULT']._serialized_start=10280
  _globals['_FILEGETRESULT']._serialized_end=10347
  _globals['_RECORDDETAILSINCLUDE']._serialized_start=10349
  _globals['_RECORDDETAILSINCLUDE']._serialized_end=10423
  _globals['_CHECKSHAREADMINOBJECTTYPE']._serialized_start=10425
  _globals['_CHECKSHAREADMINOBJECTTYPE']._serialized_end=10523
  _globals['_SHARESTATUS']._serialized_start=10525
  _globals['_SHARESTATUS']._serialized_end=10574
  _globals['_RECORDTRANSACTIONTYPE']._serialized_start=10576
  _globals['_RECORDTRANSACTIONTYPE']._serialized_end=10634
  _globals['_TIMELIMITEDACCESSTYPE']._serialized_start=10637
  _globals['_TIMELIMITEDACCESSTYPE']._serialized_end=10867
  _globals['_TIMERNOTIFICATIONTYPE']._serialized_start=10869
  _globals['_TIMERNOTIFICATIONTYPE']._serialized_end=10961
  _globals['_RECORDTYPE']._serialized_start=25
  _globals['_RECORDTYPE']._serialized_end=117
  _globals['_RECORDTYPESREQUEST']._serialized_start=119
  _globals['_RECORDTYPESREQUEST']._serialized_end=204
  _globals['_RECORDTYPESRESPONSE']._serialized_start=207
  _globals['_RECORDTYPESRESPONSE']._serialized_end=363
  _globals['_RECORDTYPEMODIFYRESPONSE']._serialized_start=365
  _globals['_RECORDTYPEMODIFYRESPONSE']._serialized_end=430
  _globals['_RECORDSGETREQUEST']._serialized_start=432
  _globals['_RECORDSGETREQUEST']._serialized_end=493
  _globals['_RECORD']._serialized_start=496
  _globals['_RECORD']._serialized_end=705
  _globals['_FOLDERRECORDKEY']._serialized_start=707
  _globals['_FOLDERRECORDKEY']._serialized_end=784
  _globals['_FOLDER']._serialized_start=786
  _globals['_FOLDER']._serialized_end=883
  _globals['_TEAM']._serialized_start=886
  _globals['_TEAM']._serialized_end=1035
  _globals['_RECORDSGETRESPONSE']._serialized_start=1038
  _globals['_RECORDSGETRESPONSE']._serialized_end=1210
  _globals['_RECORDLINK']._serialized_start=1212
  _globals['_RECORDLINK']._serialized_end=1264
  _globals['_RECORDAUDIT']._serialized_start=1266
  _globals['_RECORDAUDIT']._serialized_end=1310
  _globals['_SECURITYDATA']._serialized_start=1312
  _globals['_SECURITYDATA']._serialized_end=1340
  _globals['_SECURITYSCOREDATA']._serialized_start=1342
  _globals['_SECURITYSCOREDATA']._serialized_end=1375
  _globals['_RECORDADD']._serialized_start=1378
  _globals['_RECORDADD']._serialized_end=1766
  _globals['_RECORDSADDREQUEST']._serialized_start=1769
  _globals['_RECORDSADDREQUEST']._serialized_end=1902
  _globals['_RECORDUPDATE']._serialized_start=1905
  _globals['_RECORDUPDATE']._serialized_end=2239
  _globals['_RECORDSUPDATEREQUEST']._serialized_start=2242
  _globals['_RECORDSUPDATEREQUEST']._serialized_end=2381
  _globals['_RECORDFILEFORCONVERSION']._serialized_start=2384
  _globals['_RECORDFILEFORCONVERSION']._serialized_end=2526
  _globals['_RECORDFOLDERFORCONVERSION']._serialized_start=2528
  _globals['_RECORDFOLDERFORCONVERSION']._serialized_end=2602
  _globals['_RECORDCONVERTTOV3']._serialized_start=2605
  _globals['_RECORDCONVERTTOV3']._serialized_end=2879
  _globals['_RECORDSCONVERTTOV3REQUEST']._serialized_start=2881
  _globals['_RECORDSCONVERTTOV3REQUEST']._serialized_end=2974
  _globals['_RECORDSREMOVEREQUEST']._serialized_start=2976
  _globals['_RECORDSREMOVEREQUEST']._serialized_end=3015
  _globals['_RECORDREVERT']._serialized_start=3017
  _globals['_RECORDREVERT']._serialized_end=3079
  _globals['_RECORDSREVERTREQUEST']._serialized_start=3081
  _globals['_RECORDSREVERTREQUEST']._serialized_end=3143
  _globals['_RECORDLINKERROR']._serialized_start=3145
  _globals['_RECORDLINKERROR']._serialized_end=3244
  _globals['_RECORDMODIFYSTATUS']._serialized_start=3247
  _globals['_RECORDMODIFYSTATUS']._serialized_end=3396
  _globals['_RECORDSMODIFYRESPONSE']._serialized_start=3398
  _globals['_RECORDSMODIFYRESPONSE']._serialized_end=3485
  _globals['_RECORDADDAUDITDATA']._serialized_start=3487
  _globals['_RECORDADDAUDITDATA']._serialized_end=3576
  _globals['_ADDAUDITDATAREQUEST']._serialized_start=3578
  _globals['_ADDAUDITDATAREQUEST']._serialized_end=3645
  _globals['_FILE']._serialized_start=3647
  _globals['_FILE']._serialized_end=3763
  _globals['_FILESADDREQUEST']._serialized_start=3765
  _globals['_FILESADDREQUEST']._serialized_end=3833
  _globals['_FILEADDSTATUS']._serialized_start=3836
  _globals['_FILEADDSTATUS']._serialized_end=4003
  _globals['_FILESADDRESPONSE']._serialized_start=4005
  _globals['_FILESADDRESPONSE']._serialized_end=4080
  _globals['_FILESGETREQUEST']._serialized_start=4082
  _globals['_FILESGETREQUEST']._serialized_end=4184
  _globals['_FILEGETSTATUS']._serialized_start=4187
  _globals['_FILEGETSTATUS']._serialized_end=4349
  _globals['_FILESGETRESPONSE']._serialized_start=4351
  _globals['_FILESGETRESPONSE']._serialized_end=4408
  _globals['_APPLICATIONADDREQUEST']._serialized_start=4411
  _globals['_APPLICATIONADDREQUEST']._serialized_end=4552
  _globals['_GETRECORDDATAWITHACCESSINFOREQUEST']._serialized_start=4555
  _globals['_GETRECORDDATAWITHACCESSINFOREQUEST']._serialized_end=4691
  _globals['_USERPERMISSION']._serialized_start=4694
  _globals['_USERPERMISSION']._serialized_end=4956
  _globals['_SHAREDFOLDERPERMISSION']._serialized_start=4959
  _globals['_SHAREDFOLDERPERMISSION']._serialized_end=5175
  _globals['_RECORDDATA']._serialized_start=5178
  _globals['_RECORDDATA']._serialized_end=5538
  _globals['_RECORDDATAWITHACCESSINFO']._serialized_start=5541
  _globals['_RECORDDATAWITHACCESSINFO']._serialized_end=5741
  _globals['_GETRECORDDATAWITHACCESSINFORESPONSE']._serialized_start=5744
  _globals['_GETRECORDDATAWITHACCESSINFORESPONSE']._serialized_end=5881
  _globals['_ISOBJECTSHAREADMIN']._serialized_start=5883
  _globals['_ISOBJECTSHAREADMIN']._serialized_end=5989
  _globals['_AMISHAREADMIN']._serialized_start=5991
  _globals['_AMISHAREADMIN']._serialized_end=6063
  _globals['_RECORDSHAREUPDATEREQUEST']._serialized_start=6066
  _globals['_RECORDSHAREUPDATEREQUEST']._serialized_end=6254
  _globals['_SHAREDRECORD']._serialized_start=6257
  _globals['_SHAREDRECORD']._serialized_end=6581
  _globals['_RECORDSHAREUPDATERESPONSE']._serialized_start=6584
  _globals['_RECORDSHAREUPDATERESPONSE']._serialized_end=6797
  _globals['_SHAREDRECORDSTATUS']._serialized_start=6799
  _globals['_SHAREDRECORDSTATUS']._serialized_end=6889
  _globals['_GETRECORDPERMISSIONSREQUEST']._serialized_start=6891
  _globals['_GETRECORDPERMISSIONSREQUEST']._serialized_end=6962
  _globals['_GETRECORDPERMISSIONSRESPONSE']._serialized_start=6964
  _globals['_GETRECORDPERMISSIONSRESPONSE']._serialized_end=7048
  _globals['_RECORDPERMISSION']._serialized_start=7050
  _globals['_RECORDPERMISSION']._serialized_end=7158
  _globals['_GETSHAREOBJECTSREQUEST']._serialized_start=7160
  _globals['_GETSHAREOBJECTSREQUEST']._serialized_end=7264
  _globals['_GETSHAREOBJECTSRESPONSE']._serialized_start=7267
  _globals['_GETSHAREOBJECTSRESPONSE']._serialized_end=7626
  _globals['_SHAREUSER']._serialized_start=7629
  _globals['_SHAREUSER']._serialized_end=7794
  _globals['_SHARETEAM']._serialized_start=7796
  _globals['_SHARETEAM']._serialized_end=7864
  _globals['_SHAREENTERPRISE']._serialized_start=7866
  _globals['_SHAREENTERPRISE']._serialized_end=7929
  _globals['_RECORDSONWERSHIPTRANSFERREQUEST']._serialized_start=7931
  _globals['_RECORDSONWERSHIPTRANSFERREQUEST']._serialized_end=8014
  _globals['_TRANSFERRECORD']._serialized_start=8016
  _globals['_TRANSFERRECORD']._serialized_end=8107
  _globals['_RECORDSONWERSHIPTRANSFERRESPONSE']._serialized_start=8109
  _globals['_RECORDSONWERSHIPTRANSFERRESPONSE']._serialized_end=8204
  _globals['_TRANSFERRECORDSTATUS']._serialized_start=8206
  _globals['_TRANSFERRECORDSTATUS']._serialized_end=8298
  _globals['_RECORDSUNSHAREREQUEST']._serialized_start=8300
  _globals['_RECORDSUNSHAREREQUEST']._serialized_end=8421
  _globals['_RECORDSUNSHARERESPONSE']._serialized_start=8424
  _globals['_RECORDSUNSHARERESPONSE']._serialized_end=8558
  _globals['_RECORDSUNSHAREFOLDER']._serialized_start=8560
  _globals['_RECORDSUNSHAREFOLDER']._serialized_end=8626
  _globals['_RECORDSUNSHAREUSER']._serialized_start=8628
  _globals['_RECORDSUNSHAREUSER']._serialized_end=8687
  _globals['_RECORDSUNSHAREFOLDERSTATUS']._serialized_start=8689
  _globals['_RECORDSUNSHAREFOLDERSTATUS']._serialized_end=8761
  _globals['_RECORDSUNSHAREUSERSTATUS']._serialized_start=8763
  _globals['_RECORDSUNSHAREUSERSTATUS']._serialized_end=8828
  _globals['_TIMEDACCESSCALLBACKPAYLOAD']._serialized_start=8830
  _globals['_TIMEDACCESSCALLBACKPAYLOAD']._serialized_end=8921
  _globals['_TIMELIMITEDACCESSREQUEST']._serialized_start=8924
  _globals['_TIMELIMITEDACCESSREQUEST']._serialized_end=9177
  _globals['_TIMELIMITEDACCESSSTATUS']._serialized_start=9179
  _globals['_TIMELIMITEDACCESSSTATUS']._serialized_end=9234
  _globals['_TIMELIMITEDACCESSRESPONSE']._serialized_start=9237
  _globals['_TIMELIMITEDACCESSRESPONSE']._serialized_end=9464
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
import logging
import argparse
import json
import sys
import os.path

from keeper_secrets_manager_core.utils import url_safe_str_to_bytes
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ..pam.router_helper import router_get_connected_gateways, router_set_record_rotation_information
from ... import api, subfolder, utils, crypto, vault, vault_extensions
from ...display import bcolors
from ...proto import router_pb2, record_pb2
from ...discovery_common.jobs import Jobs
from ...discovery_common.process import Process, QuitException, NoDiscoveryDataException
from ...discovery_common.types import (
    DiscoveryObject, UserAcl, PromptActionEnum, PromptResult, BulkRecordAdd, BulkRecordConvert, BulkProcessResults,
    BulkRecordSuccess, BulkRecordFail, DirectoryInfo, NormalizedRecord, RecordField)
from pydantic import BaseModel
from typing import Optional, List, Any, TYPE_CHECKING

from ...api import get_records_add_request

if TYPE_CHECKING:
    from ...params import KeeperParams
    from ...vault import TypedRecord, KeeperRecord
    from ...keeper_dag.vertex import DAGVertex
    from ...discovery_common.record_link import RecordLink


def _h(value: str) -> str:
    return f"{bcolors.HEADER}{value}{bcolors.ENDC}"


def _b(value: str) -> str:
    return f"{bcolors.BOLD}{value}{bcolors.ENDC}"


def _f(value: str) -> str:
    return f"{bcolors.FAIL}{value}{bcolors.ENDC}"


def _ok(value: str) -> str:
    return f"{bcolors.OKGREEN}{value}{bcolors.ENDC}"


# This is used for the admin user search
class AdminSearchResult(BaseModel):
    record: Any
    is_directory_user: bool
    is_pam_user: bool
    being_used: bool = False


class PAMGatewayActionDiscoverResultProcessCommand(PAMGatewayActionDiscoverCommandBase):

    """
    Process the discovery data
    """

    parser = argparse.ArgumentParser(prog='pam-action-discover-process')
    parser.add_argument('--job-id', '-j', required=True, dest='job_id', action='store',
                        help='Discovery job to process.')

    # This is not ready yet.
    # parser.add_argument('--smart-add', required=False, dest='smart_add', action='store_true',
    #                     help='Automatically add resources with credentials and their users.')

    parser.add_argument('--add-all', required=False, dest='add_all', action='store_true',
                        help='Respond with ADD for all prompts.')
    parser.add_argument('--debug-gs-level', required=False, dest='debug_level', action='store',
                        help='GraphSync debug level. Default is 0', type=int, default=0)

    EDITABLE = [
        "login",
        "password",
        "distinguishedName",
        "alternativeIPs",
        "database",
        "privatePEMKey",
        "connectDatabase",
        "operatingSystem"
    ]

    def get_parser(self):
        return PAMGatewayActionDiscoverResultProcessCommand.parser

    @staticmethod
    def _is_directory_user(record_type: str) -> bool:
        # pamAzureConfiguration has tenant users what are like a directory.
        return (record_type == "pamDirectory" or
                record_type == "pamAzureConfiguration")

    @staticmethod
    def _get_shared_folder(params: KeeperParams, pad: str, gateway_context: GatewayContext) -> str:
        while True:
            shared_folders = gateway_context.get_shared_folders(params)
            index = 0
            for folder in shared_folders:
                print(f"{pad}* {_h(str(index+1))} - {folder.get('uid')}  {folder.get('name')}")
                index += 1
            selected = input(f"{pad}Enter number of the shared folder>")
            try:
                return shared_folders[int(selected) - 1].get("uid")
            except ValueError:
                print(f"{pad}{_f('Input was not a number.')}")

    @staticmethod
    def get_field_values(record: TypedRecord, field_type: str) -> List[str]:
        return next(
            (f.value
             for f in record.fields
             if f.type == field_type),
            None
        )

    def get_keys_by_record(self, params: KeeperParams, gateway_context: GatewayContext,
                           record: TypedRecord) -> List[str]:
        """
        For the record, get the values of fields that are key for this record type.

        :param params:
        :param gateway_context:
        :param record:
        :return:
        """

        key_field = Process.get_key_field(record.record_type)
        keys = []
        if key_field == "host_port":
            values = self.get_field_values(record, "pamHostname")
            if len(values) == 0:
                return []

            host = values[0].get("hostName")
            port = values[0].get("port")
            if port is not None:
                if host is not None:
                    keys.append(f"{host}:{port}".lower())

        elif key_field == "host":
            values = self.get_field_values(record, "pamHostname")
            if len(values) == 0:
                return []

            host = values[0].get("hostName")
            if host is not None:
                keys.append(host.lower())

        elif key_field == "user":

            # This is user protobuf values.
            # We could make this also use record linking if we stop using protobuf.

            record_rotation = params.record_rotation_cache.get(record.record_uid)
            if record_rotation is not None:
                controller_uid = record_rotation.get("configuration_uid")
                if controller_uid is None or controller_uid != gateway_context.configuration_uid:
                    return []

                resource_uid = record_rotation.get("resource_uid")
                # If the resource uid is None, the Admin Cred Record has not been set.
                if resource_uid is None:
                    return []

                values = self.get_field_values(record, "login")
                if len(values) == 0:
                    return []

                keys.append(f"{resource_uid}:{values[0]}".lower())

        return keys

    @staticmethod
    def _record_lookup(record_uid: str,  context: Optional[Any] = None) -> Optional[NormalizedRecord]:

        """
        Get the record from the Vault, normalize it, and return it.

        Since common code is using this method we want to flatten/abstract the KeeperRecord/TypedRecord.
        """

        params = context.get("params")
        record = vault.TypedRecord.load(params, record_uid)  # type: Optional[TypedRecord]
        if record is None:
            return None

        normalized_record = NormalizedRecord(
            record_uid=record.record_uid,
            record_type=record.record_type,
            title=record.title,
            notes=record.notes
        )
        for field in record.fields:
            normalized_record.fields.append(
                RecordField(
                    type=field.type,
                    label=field.label,
                    value=field.value,
                )
            )
        if record.custom is not None:
            for field in record.custom:
                normalized_record.fields.append(
                    RecordField(
                        type=field.type,
                        label=field.label,
                        value=field.value,
                    )
                )
        return normalized_record

    def _build_record_cache(self, params: KeeperParams, gateway_context: GatewayContext) -> dict:

        """
        Make a lookup cache for all the records.

        This is used to flag discovered items as existing if the record has already been added. This is used to
        prevent duplicate records being added.
        """

        logging.debug(f"building the PAM record cache")

        # Make a cache of existing record by the criteria per record type
        cache = {
            "pamUser": {},
            "pamMachine": {},
            "pamDirectory": {},
            "pamDatabase": {}
        }

        # Set all the PAM Records
        records = list(vault_extensions.find_records(params, "pam*"))
        for record in records:
            # If the record type is not part of the cache, skip the record
            if record.record_type not in cache:
                continue

            # Load the full record
            record = vault.TypedRecord.load(params, record.record_uid)  # type: Optional[TypedRecord]

            cache_keys = self.get_keys_by_record(
                params=params,
                gateway_context=gateway_context,
                record=record
            )
            if len(cache_keys) == 0:
                continue

            for cache_key in cache_keys:
                cache[record.record_type][cache_key] = record.record_uid

        return cache

    def _edit_record(self, content: DiscoveryObject, pad: str, editable: List[str]) -> bool:

        edit_label = input(f"{pad}Enter 'title' or the name of the {_ok('Label')} to edit, RETURN to cancel> ")

        # Just pressing return exits the edit
        if edit_label == "":
            return False

        # If the "title" is entered, then edit the title of the record.
        if edit_label.lower() == "title":
            new_title = input(f"{pad}Enter new title> ")
            content.title = new_title

        # If a field label is entered, and it's in the list of editable fields, then allow the user to edit.
        elif edit_label in editable:
            new_value = None
            if edit_label in self.FIELD_MAPPING:
                type_hint = self.FIELD_MAPPING[edit_label].get("type")
                if type_hint == "dict":
                    field_input_format = self.FIELD_MAPPING[edit_label].get("field_input")
                    new_value = {}
                    for field in field_input_format:
                        new_value[field.get('key')] = input(f"{pad}Enter {field_input_format.get('prompt')} value> ")
                elif type_hint == "csv":
                    new_value = input(f"{pad}Enter {edit_label} values, separate with a comma > ")
                    new_values = map(str.strip, new_value.split(','))
                    new_value = "\n".join(new_values)
                elif type_hint == "multiline":
                    print(_b(f"{pad}Enter multilines of text or a path, on the first line, "
                             "to a file that contains the value."))
                    print(_b(f"{pad}To end, type 'END' at the start of a new line. You can paste text."))
                    new_value = ""
                    first_line = True
                    while True:
                        line = input(_b(f"> ")).rstrip()
                        if line == "END":
                            break

                        # If this is the first line, check if line is a path to a file.
                        if first_line is True:
                            try:
                                test_file = line.strip()
                                logging.debug(f"is first line, check for file path for '{test_file}'")
                                if os.path.exists(test_file) is True:
                                    with open(test_file, "r") as fh:
                                        new_value = fh.read()
                                        fh.close()
                                        break
                                else:
                                    logging.debug(f"first line is not a file path")
                            except Exception as err:
                                logging.debug(f"exception checking if file: {err}")
                        first_line = False
                        new_value += line + "\n"
                elif type_hint == "choice":

                    values = self.FIELD_MAPPING[edit_label].get("values")
                    text_values = [_b(x) for x in values]
                    new_value = input(f"{pad}Enter one of the follow values: {', '.join(text_values)}> ")
                    new_value = new_value.strip().lower()
                    if new_value not in values:
                        print(f"{pad}{_f('The value ' + new_value + ' is not one of the values allowed.')}")
                        return False
            else:
                new_value = input(f"{pad}Enter new value, or path to a file that contains the value > ")

                # Is the value a path to a file, i.e., a private key file.
                try:
                    if os.path.exists(new_value) is True:
                        with open(new_value, "r") as fh:
                            new_value = fh.read()
                            fh.close()
                except (Exception,):
                    pass

            for edit_field in content.fields:
                if edit_field.label == edit_label:
                    edit_field.value = [new_value]

        # Else, the label they entered cannot be edited.
        else:
            print(f"{pad}{_f('The field is not editable.')}")
            return False

        return True

    @staticmethod
    def _add_all_preprocess(vertex: DAGVertex, content: DiscoveryObject, parent_vertex: DAGVertex,
                            acl: Optional[UserAcl] = None) -> Optional[PromptResult]:
        """
        This is client side check if we should skip prompting the user.

        The checks are
        * A directory with the same domain already has a record.

        """

        _ = vertex
        _ = acl

        # Check if the directory for a domain exists.
        # From the parent, find any directory objects.
        # If they already have a record UID, don't prompt about this one.
        # Once a directory for the domain exists, the user should not be prompted about this domain anymore.
        if content.record_type == "pamDirectory":
            for v in parent_vertex.has_vertices():
                other_content = DiscoveryObject.get_discovery_object(v)
                if other_content.record_uid is not None and other_content.name == content.name:
                    return PromptResult(action=PromptActionEnum.SKIP)
        return None

    def _prompt_display_fields(self, content: DiscoveryObject, pad: str) -> List[str]:

        editable = []
        for field in content.fields:
            has_editable = False
            if field.label in PAMGatewayActionDiscoverResultProcessCommand.EDITABLE:
                editable.append(field.label)
                has_editable = True
            value = field.value

            # If there is a value, and it's not just [], also make sure the
            if len(value) > 0 and value[0] is not None:
                # PAM records will have only 1 item in the value array.
                value = value[0]
                if field.label in self.FIELD_MAPPING:
                    type_hint = self.FIELD_MAPPING[field.label].get("type")
                    formatted_value = []
                    if type_hint == "dict":
                        field_input_format = self.FIELD_MAPPING[field.label].get("field_format")
                        for format_field in field_input_format:
                            formatted_value.append(f"{format_field.get('label')}: "
                                                   f"{value.get(format_field.get('key'))}")
                    elif type_hint == "csv":
                        formatted_value.append(", ".join(value.split("\n")))
                    elif type_hint == "multiline":
                        formatted_value.append(value)
                    elif type_hint == "choice":
                        formatted_value.append(value)
                    value = ", ".join(formatted_value)
            else:
                if has_editable is True:
                    value = f"{bcolors.FAIL}MISSING{bcolors.ENDC}"
                else:
                    value = f"{bcolors.OKBLUE}None{bcolors.ENDC}"

            color = bcolors.HEADER
            if has_editable is True:
                color = bcolors.OKGREEN

            rows = str(value).split("\n")
            if len(rows) > 1:
                value = rows[0] + _b(f"... {len(rows)} rows.")

            print(f"{pad}  "
                  f"{color}Label:{bcolors.ENDC} {field.label}, "
                  f"{_h('Type:')} {field.type}, "
                  f"{_h('Value:')} {value}")

        if len(content.notes) > 0:
            print("")
            for note in content.notes:
                print(f"{pad}* {note}")

        return editable

    @staticmethod
    def _prompt_display_relationships(vertex: DAGVertex, content: DiscoveryObject, pad: str):

        if vertex is None:
            return

        if content.record_type == "pamUser":
            belongs_to = []
            for v in vertex.belongs_to_vertices():
                resource_content = DiscoveryObject.get_discovery_object(v)
                belongs_to.append(resource_content.name)
            count = len(belongs_to)
            print("")
            print(f"{pad}This user is found on {count} resource{'s' if count > 1 else ''}")

    def _prompt(self,
                content: DiscoveryObject,
                acl: UserAcl,
                vertex: Optional[DAGVertex] = None,
                parent_vertex: Optional[DAGVertex] = None,
                resource_has_admin: bool = True,
                item_count: int = 0,
                items_left: int = 0,
                indent: int = 0,
                block_auto_add: bool = False,
                context: Optional[Any] = None) -> PromptResult:

        if context is None:
            raise Exception("Context not set for processing the discovery results")

        parent_content = DiscoveryObject.get_discovery_object(parent_vertex)

        print("")

        params = context.get("params")
        gateway_context = context.get("gateway_context")
        dry_run = context.get("dry_run", False)
        add_all = context.get("add_all", False)
        if block_auto_add:
            add_all = False

        # If auto add is True, there are sometime we don't want to add the object.
        # If we get a result, we want to return it.
        # Skip the prompt.
        if add_all is True and vertex is not None:
            result = self._add_all_preprocess(vertex, content, parent_vertex, acl)
            if result is not None:
                return result

        # If the record type is a pamUser, then include parent description.
        if content.record_type == "pamUser" and parent_vertex is not None:
            parent_pad = ""
            if indent - 1 > 0:
                parent_pad = "".ljust(2 * indent, ' ')

            print(f"{parent_pad}{_h(parent_content.description)}")

        pad = ""
        if indent > 0:
            pad = "".ljust(2 * indent, ' ')

        print(f"{pad}{_h(content.description)}")

        show_current_object = True
        while show_current_object is True:
            print(f"{pad}{bcolors.HEADER}Record Title:{bcolors.ENDC} {content.title}")

            logging.debug(f"Fields: {content.fields}")

            # Display the fields and return a list of fields are editable.
            editable = self._prompt_display_fields(content=content, pad=pad)
            if vertex is not None:
                self._prompt_display_relationships(vertex=vertex, content=content, pad=pad)

            while True:

                shared_folder_uid = content.shared_folder_uid
                if shared_folder_uid is None:
                    shared_folder_uid = gateway_context.default_shared_folder_uid

                count_prompt = ""
                if item_count > 0:
                    count_prompt = f"{bcolors.HEADER}[{item_count - items_left + 1}/{item_count}]{bcolors.ENDC}"
                edit_add_prompt = f"{count_prompt} "
                if len(editable) > 0:
                    edit_add_prompt += f"({_b('E')})dit, "

                shared_folders = gateway_context.get_shared_folders(params)
                if dry_run is False:
                    if len(shared_folders) > 1:
                        folder_name = next((x['name']
                                            for x in shared_folders
                                            if x['uid'] == shared_folder_uid),
                                           None)
                        edit_add_prompt += f"({_b('A')})dd to {folder_name}, "\
                                           f"Add to ({_b('F')})older, "
                    else:
                        if dry_run is False:
                            edit_add_prompt += f"({_b('A')})dd, "
                prompt = f"{edit_add_prompt}({_b('S')})kip, ({_b('I')})gnore, ({_b('Q')})uit"

                command = "a"
                if add_all is False:
                    command = input(f"{pad}{prompt}> ").lower()
                if (command == "a" or command == "f") and dry_run is False:

                    print(f"{pad}{bcolors.OKGREEN}Adding record to save queue.{bcolors.ENDC}")
                    print("")

                    if command == "f":
                        shared_folder_uid = self._get_shared_folder(params, pad, gateway_context)

                    content.shared_folder_uid = shared_folder_uid

                    # This happens when the record is a pamUser and parent resource record does not have an
                    #   administrator.
                    # It's like the reverse of creating an admin after adding the resource.
                    # It would make this user the admin for the parent resource.
                    # This condition would be really rare, since to get the users, the resource would have to have an
                    #  admin user.
                    if content.record_type == "pamUser" and resource_has_admin is False:

                        print(_b(f"{parent_content.description} does not have an administrator."))
                        if (hasattr(parent_content.item, "admin_reason") and
                                parent_content.item.admin_reason is not None):
                            print("")
                            print(parent_content.item.admin_reason)
                            print("")

                        while True:

                            yn = input("Do you want to make this user the administrator? [Y/N]> ").lower()
                            if yn == "":
                                continue
                            if yn[0] == "n":
                                break
                            if yn[0] == "y":
                                acl.is_admin = True
                                break

                    return PromptResult(
                        action=PromptActionEnum.ADD,
                        acl=acl,
                        content=content
                    )

                elif command == "e" and dry_run is False:
                    self._edit_record(content, pad, editable)
                    break

                elif command == "i":

                    print(f"{pad}{bcolors.OKBLUE}Creating an ignore rule for record.{bcolors.ENDC}")
                    return PromptResult(
                        action=PromptActionEnum.IGNORE,
                        acl=acl,
                        content=content
                    )

                elif command == "s":
                    print(f"{pad}{bcolors.OKBLUE}Skipping record.{bcolors.ENDC}")

                    return PromptResult(
                        action=PromptActionEnum.SKIP,
                        acl=acl,
                        content=content
                    )
                elif command == "q":
                    raise QuitException()
            print()

    def _find_user_record(self,
                          params: KeeperParams,
                          bulk_convert_records: List[BulkRecordConvert],
                          context: Optional[Any] = None) -> (Optional[TypedRecord], bool):

        gateway_context = context.get("gateway_context")  # type: GatewayContext
        record_link = context.get("record_link")  # type: RecordLink

        # Get the latest records
        params.sync_data = True

        # Make a list of all records in the shared folders.
        # We will use this to check if a selected user is in the shared folders.
        shared_record_uids = []
        for shared_folder in gateway_context.get_shared_folders(params):
            folder = shared_folder.get("folder")
            if "records" in folder:
                for record in folder["records"]:
                    shared_record_uids.append(record.get("record_uid"))

        # Make a list of record we are already converting so we don't show them again.
        converting_list = [x.record_uid for x in bulk_convert_records]

        logging.debug(f"shared folders record uid {shared_record_uids}")

        while True:
            user_search = input("Enter an user to search for [ENTER/RETURN to quit]> ")
            if user_search == "":
                print(f"{bcolors.FAIL}No search terms, not performing search.{bcolors.ENDC}")
                return None, False

            # Search for record with the search string.
            # Currently, this only works with TypedRecord, version 3.
            user_record = list(vault_extensions.find_records(
                params,
                search_str=user_search,
                record_version=3
            ))
            # If not record are returned by the search just return None,
            if len(user_record) == 0:
                print(f"{bcolors.FAIL}Could not find any records that contain the search text.{bcolors.ENDC}")
                return None, False

            # Find usable admin records.
            admin_search_results = []  # type: List[AdminSearchResult]
            for record in user_record:

                user_record = vault.KeeperRecord.load(params, record.record_uid)
                if user_record.record_type == "pamUser":
                    logging.debug(f"{record.record_uid} is a pamUser")

                    # If we are already converting this pamUser record, then don't show it.
                    if record.record_uid in converting_list:
                        logging.debug(f"pamUser {user_record.title}, {user_record.record_uid} is being converted; "
                                      "BAD for search")
                        admin_search_results.append(
                            AdminSearchResult(
                                record=user_record,
                                is_directory_user=False,
                                is_pam_user=True,
                                being_used=True
                            )
                        )
                        continue

                    # Does the record exist in the gateway shared folder?
                    # We want to filter our other gateway's pamUser, or it will get overwhelming.
                    if user_record.record_uid not in shared_record_uids:
                        logging.debug(f"pamUser {record.title}, {user_record.record_uid} not in shared "
                                      "folder, BAD for search")
                        continue

                    # If the record does not exist in the record linking, it's orphaned; accept it
                    # If it does exist, then check if it belonged to a directory.
                    # Very unlikely a user that belongs to a database or another machine can be used.

                    record_vertex = record_link.get_record_link(user_record.record_uid)
                    is_directory_user = False
                    if record_vertex is not None:
                        parent_record_uid = record_link.get_parent_record_uid(user_record.record_uid)
                        parent_record = vault.TypedRecord.load(params, parent_record_uid)  # type: Optional[TypedRecord]
                        if parent_record is not None:
                            is_directory_user = self._is_directory_user(parent_record.record_type)
                            if is_directory_user is False:
                                logging.debug(f"pamUser parent for {user_record.title}, "
                                              "{user_record.record_uid} is not a directory; BAD for search")
                                continue

                            logging.debug(f"pamUser {user_record.title}, {user_record.record_uid} is a directory user; "
                                          "good for search")

                        else:
                            logging.debug(f"pamUser {user_record.title}, {user_record.record_uid} does not a parent; "
                                          "good for search")
                    else:
                        logging.debug(f"pamUser {user_record.title}, {user_record.record_uid} does not have record "
                                      "linking vertex; good for search")

                    admin_search_results.append(
                        AdminSearchResult(
                            record=user_record,
                            is_directory_user=is_directory_user,
                            is_pam_user=True,
                            being_used=False
                        )
                    )

                # Else this is a non-PAM record.
                # Make sure it has a login, password, private key
                else:
                    logging.debug(f"{record.record_uid} is NOT a pamUser")
                    login_field = next((x for x in record.fields if x.type == "login"), None)
                    password_field = next((x for x in record.fields if x.type == "password"), None)
                    private_key_field = next((x for x in record.fields if x.type == "keyPair"), None)

                    if login_field is not None and (password_field is not None or private_key_field is not None):
                        admin_search_results.append(
                            AdminSearchResult(
                                record=record,
                                is_directory_user=False,
                                is_pam_user=False
                            )
                        )
                        logging.debug(f"{record.title} is has credentials, good for search")
                    else:
                        logging.debug(f"{record.title} is missing full credentials, BAD for search")

            # If all the users have been filtered out, then just return None
            if len(admin_search_results) == 0:
                print(f"{bcolors.FAIL}Could not find any available records.{bcolors.ENDC}")
                return None, False

            user_index = 1
            admin_search_results = sorted(admin_search_results,
                                          key=lambda x: x.is_pam_user,
                                          reverse=True)

            has_local_user = False
            for admin_search_result in admin_search_results:
                is_local_user = False
                if admin_search_result.record.record_type != "pamUser":
                    has_local_user = True
                    is_local_user = True

                hc = bcolors.HEADER
                b = bcolors.BOLD
                tc = ""
                index_str = user_index
                if admin_search_result.being_used is True:
                    hc = bcolors.WARNING
                    b = bcolors.WARNING
                    tc = bcolors.WARNING
                    index_str = "-" * len(str(index_str))

                print(f"{hc}[{index_str}] {bcolors.ENDC}"
                      f"{b + '* ' + bcolors.ENDC if is_local_user is True else ''}"
                      f"{tc}{admin_search_result.record.title}{bcolors.ENDC} "
                      f'{"(Directory User) " if admin_search_result.is_directory_user is True else ""}'
                      f'{tc + "(Already taken)" + bcolors.ENDC if admin_search_result.being_used is True else ""}')
                user_index += 1

            if has_local_user is True:
                print(f"{bcolors.BOLD}* Not a PAM User record. "
                      f"A PAM User would be generated from this record.{bcolors.ENDC}")

            select = input("Enter line number of user record to use, enter/return to refine the search, "
                           f"or {_b('Q')} to quit search. > ").lower()
            if select == "":
                continue
            elif select[0] == "q":
                return None, False
            else:
                try:
                    selected = admin_search_results[int(select) - 1]
                    if selected.being_used is True:
                        print(f"{bcolors.FAIL}Cannot select a record that has already been taken. "
                              f"Another record is using this local user as its administrator.{bcolors.ENDC}")
                        return None, False
                    admin_record = selected.record  # type: TypedRecord
                    return admin_record, selected.is_directory_user
                except IndexError:
                    print(f"{bcolors.FAIL}Entered row index does not exists.{bcolors.ENDC}")
                    continue

    @staticmethod
    def _handle_admin_record_from_record(record: TypedRecord,
                                         content: DiscoveryObject,
                                         context: Optional[Any] = None) -> Optional[PromptResult]:

        params = context.get("param")  # type: KeeperParams
        gateway_context = context.get("gateway_context")  # type: GatewayContext

        # Is this a pamUser record?
        # Return the record UID and set its ACL to be the admin.
        if record.record_type == "pamUser":
            return PromptResult(
                action=PromptActionEnum.ADD,
                acl=UserAcl(is_admin=True),
                record_uid=record.record_uid,
            )

        # If we are here, this was not a pamUser
        # We need to duplicate the record.
        # But confirm first

        # Get fields from the old record.
        # Copy them into the fields.
        login_field = next((x for x in record.fields if x.type == "login"), None)
        password_field = next((x for x in record.fields if x.type == "password"), None)
        private_key_field = next((x for x in record.fields if x.type == "keyPair"), None)

        content.set_field_value("login", login_field.value)
        if password_field is not None:
            content.set_field_value("password", password_field.value)
        if private_key_field is not None:
            value = private_key_field.value
            if value is not None and len(value) > 0:
                value = value[0]
                private_key = value.get("privateKey")
                if private_key is not None:
                    content.set_field_value("private_key", private_key)

        # Check if we have more than one shared folder.
        # If we have one, confirm about adding the user.
        # If multiple shared folders, allow user to select which one.
        shared_folders = gateway_context.get_shared_folders(params)
        if len(shared_folders) == 0:
            while True:
                yn = input(f"Create a PAM User record from {record.title}? [Y/N]> ").lower()
                if yn == "":
                    continue
                elif yn[0] == "n":
                    return None
                elif yn[0] == "y":
                    content.shared_folder_uid = gateway_context.default_shared_folder_uid
        else:
            folder_name = next((x['name']
                                for x in shared_folders
                                if x['uid'] == gateway_context.default_shared_folder_uid),
                               None)
            while True:
                shared_folders = gateway_context.get_shared_folders(params)
                if len(shared_folders) > 1:
                    afq = input(f"({_b('A')})dd user to {folder_name}, "
                                f"Add user to ({_b('F')})older, "
                                f"({_b('Q')})uit > ").lower()
                else:
                    afq = input(f"({_b('A')})dd user, "
                                f"({_b('Q')})uit > ").lower()

                if afq == "":
                    continue
                if afq[0] == "a":
                    content.shared_folder_uid = gateway_context.default_shared_folder_uid
                    break
                elif afq[0] == "f":
                    shared_folder_uid = PAMGatewayActionDiscoverResultProcessCommand._get_shared_folder(
                        params, "", gateway_context)
                    if shared_folder_uid is not None:
                        content.shared_folder_uid = shared_folder_uid
                        break

        return PromptResult(
            action=PromptActionEnum.ADD,
            acl=UserAcl(is_admin=True),
            content=content,
            note=f"This record replaces record {record.title} ({record.record_uid}). "
                 "The password on that record will not be rotated."
        )

    def _prompt_admin(self,
                      parent_vertex: DAGVertex,
                      content: DiscoveryObject,
                      acl: UserAcl,
                      bulk_convert_records: List[BulkRecordConvert],
                      indent: int = 0,
                      context: Optional[Any] = None) -> PromptResult:

        if content is None:
            raise Exception("The admin content was not passed in to prompt the user.")

        params = context.get("params")

        parent_content = DiscoveryObject.get_discovery_object(parent_vertex)

        print("")
        while True:

            print(f"{bcolors.BOLD}{parent_content.description} does not have an administrator user.{bcolors.ENDC}")
            if hasattr(parent_content.item, "admin_reason") is True and parent_content.item.admin_reason is not None:
                print("")
                print(parent_content.item.admin_reason)
            print("")

            action = input("Would you like to "
                           f"({_b('A')})dd new administrator user, "
                           f"({_b('F')})ind an existing admin, or "
                           f"({_b('S')})kip add? > ").lower()

            if action == "":
                continue

            if action[0] == 'a':
                prompt_result = self._prompt(
                    vertex=None,
                    parent_vertex=parent_vertex,
                    content=content,
                    acl=acl,
                    context=context,
                    indent=indent + 2,
                    block_auto_add=True
                )
                login = content.get_field_value("login")
                if login is None or login == "":
                    print("")
                    print(f"{bcolors.FAIL}A value is needed for the login field.{bcolors.ENDC}")
                    continue

                print(f"{bcolors.OKGREEN}Adding admin record to save queue.{bcolors.ENDC}")
                return prompt_result
            elif action[0] == 'f':
                print("")
                record, is_directory_user = self._find_user_record(params,
                                                                   context=context,
                                                                   bulk_convert_records=bulk_convert_records)
                if record is not None:
                    admin_prompt_result = self._handle_admin_record_from_record(
                        record=record,
                        content=content,
                        context=context
                    )
                    if admin_prompt_result is not None:
                        if admin_prompt_result.action == PromptActionEnum.ADD:
                            admin_prompt_result.is_directory_user = is_directory_user
                            print(f"{bcolors.OKGREEN}Adding admin record to save queue.{bcolors.ENDC}")
                        return admin_prompt_result
            elif action[0] == 's':
                return PromptResult(
                    action=PromptActionEnum.SKIP
                )
            print("")

    @staticmethod
    def _display_auto_add_results(bulk_add_records: List[BulkRecordAdd]):

        """
        Display the number of record created from rule engine ADD results and smart add function.
        """

        add_count = len(bulk_add_records)
        if add_count > 0:
            print("")
            print(f"{bcolors.OKGREEN}From the rules, automatically queued {add_count} "
                  f"record{'' if add_count == 1 else 's'} to be added.{bcolors.ENDC}")

    @staticmethod
    def _prompt_confirm_add(bulk_add_records: List[BulkRecordAdd]):

        """
        If we quit, we want to ask the user if they want to add record for discovery objects that they selected
        for addition.
        """

        print("")
        count = len(bulk_add_records)
        if count == 1:
            msg = (f"{bcolors.BOLD}There is 1 record queued to be added to your vault. "
                   f"Do you wish to add it? [Y/N]> {bcolors.ENDC}")
        else:
            msg = (f"{bcolors.BOLD}There are {count} records queued to be added to your vault. "
                   f"Do you wish to add them? [Y/N]> {bcolors.ENDC}")
        while True:
            yn = input(msg).lower()
            if yn == "":
                continue
            if yn[0] == "y":
                return True
            elif yn[0] == "n":
                return False
            print(f"{bcolors.FAIL}Did not get 'Y' or 'N'{bcolors.ENDC}")

    @staticmethod
    def _prepare_record(content: DiscoveryObject, context: Optional[Any] = None) -> (Any, str):

        """
        Prepare the Vault record side.

        It's not created here.
        It will be created at the end of the processing run in bulk.
        We to build a record to get a record UID.

        :params content: The discovery object instance.
        :params context: Optionally, it will contain information set from the run() method.
        :returns: Returns an unsaved Keeper record instance.
        """

        params = context.get("params")

        # DEFINE V3 RECORD

        # Create an instance of a vault record to structure the data
        record = vault.TypedRecord()
        record.type_name = content.record_type
        record.record_uid = utils.generate_uid()
        record.record_key = utils.generate_aes_key()
        record.title = content.title
        for field in content.fields:
            field_args = {
                "field_type": field.type,
                "field_value": field.value
            }
            if field.type != field.label:
                field_args["field_label"] = field.label
            record_field = vault.TypedField.new_field(**field_args)
            record_field.required = field.required
            record.fields.append(record_field)

        folder = params.folder_cache.get(content.shared_folder_uid)
        folder_key = None  # type: Optional[bytes]
        if isinstance(folder, subfolder.SharedFolderFolderNode):
            shared_folder_uid = folder.shared_folder_uid
        elif isinstance(folder, subfolder.SharedFolderNode):
            shared_folder_uid = folder.uid
        else:
            shared_folder_uid = None
        if shared_folder_uid and shared_folder_uid in params.shared_folder_cache:
            shared_folder = params.shared_folder_cache.get(shared_folder_uid)
            folder_key = shared_folder.get('shared_folder_key_unencrypted')

        # DEFINE PROTOBUF FOR RECORD

        record_add_protobuf = record_pb2.RecordAdd()
        record_add_protobuf.record_uid = utils.base64_url_decode(record.record_uid)
        record_add_protobuf.record_key = crypto.encrypt_aes_v2(record.record_key, params.data_key)
        record_add_protobuf.client_modified_time = utils.current_milli_time()
        record_add_protobuf.folder_type = record_pb2.user_folder
        if folder:
            record_add_protobuf.folder_uid = utils.base64_url_decode(folder.uid)
            if folder.type == 'shared_folder':
                record_add_protobuf.folder_type = record_pb2.shared_folder
            elif folder.type == 'shared_folder_folder':
                record_add_protobuf.folder_type = record_pb2.shared_folder_folder
            if folder_key:
                record_add_protobuf.folder_key = crypto.encrypt_aes_v2(record.record_key, folder_key)

        data = vault_extensions.extract_typed_record_data(record)
        json_data = api.get_record_data_json_bytes(data)
        record_add_protobuf.data = crypto.encrypt_aes_v2(json_data, record.record_key)

        if params.enterprise_ec_key:
            audit_data = vault_extensions.extract_audit_data(record)
            if audit_data:
                record_add_protobuf.audit.version = 0
                record_add_protobuf.audit.data = crypto.encrypt_ec(
                    json.dumps(audit_data).encode('utf-8'), params.enterprise_ec_key)

        return record_add_protobuf, record.record_uid

    @classmethod
    def _create_records(cls, bulk_add_records: List[BulkRecordAdd], context: Optional[Any] = None) -> (
            BulkProcessResults):

        if len(bulk_add_records) == 1:
            print("Adding the record to the Vault ...")
        else:
            print(f"Adding {len(bulk_add_records)} records to the Vault ...")

        params = context.get("params")
        gateway_context = context.get("gateway_context")

        build_process_results = BulkProcessResults()

        # STEP 1 - Batch add new records

        # Generate a list of RecordAdd instance.
        # In BulkRecordAdd they will be the record instance.
        record_add_list = [r.record for r in bulk_add_records]  # type: List[record_pb2.RecordAdd]

        records_per_request = 999

        add_results = []  # type: List[record_pb2.RecordModifyResult]
        logging.debug("adding record in batches")
        print("batch record create: ", end="")
        sys.stdout.flush()
        while record_add_list:
            print(".", end="")
            sys.stdout.flush()
            logging.debug(f"* adding batch")
            rq = get_records_add_request(params)
            rq.records.extend(record_add_list[:records_per_request])
            record_add_list = record_add_list[records_per_request:]
            rs = api.communicate_rest(params, rq, 'vault/records_add', rs_type=record_pb2.RecordsModifyResponse)
            add_results.extend(rs.records)
        print("")
        sys.stdout.flush()

        logging.debug(f"add_result: {add_results}")

        if len(add_results) != len(bulk_add_records):
            logging.debug(f"attempted to batch add {len(bulk_add_records)} record(s), "
                          f"only have {len(add_results)} results.")

        # STEP 3 - Add rotation settings.
        # Use the list we passed in, find the results, and add if the additions were successful.

        # Keep track of each record we create a rotation for to avoid version problems, if there was a dup.
        created_cache = []

        # For the records passed in to be created.
        print("add rotation settings: ", end="")
        sys.stdout.flush()
        for bulk_record in bulk_add_records:
            if bulk_record.record_uid in created_cache:
                logging.debug(f"found a duplicate of record uid: {bulk_record.record_uid}")
                continue
            print(".", end="")
            sys.stdout.flush()

            # Grab the type Keeper record instance, and title from that record.
            pb_add_record = bulk_record.record
            title = bulk_record.title

            rotation_disabled = False

            # Find the result for this record.
            result = None
            for x in add_results:
                logging.debug(f"{pb_add_record.record_uid} vs {x.record_uid}")
                if pb_add_record.record_uid == x.record_uid:
                    result = x
                    break

            # If we didn't get a result, then don't add the rotation settings.
            if result is None:
                build_process_results.failure.append(
                    BulkRecordFail(
                        title=title,
                        error="No status on addition to Vault. Cannot determine if added or not."
                    )
                )
                logging.debug(f"Did not get a result when adding record {title}")
                continue

            # Check if addition failed. If it did fail, don't add the rotation settings.
            success = (result.status == record_pb2.RecordModifyResult.DESCRIPTOR.values_by_name['RS_SUCCESS'].number)
            status = record_pb2.RecordModifyResult.DESCRIPTOR.values_by_number[result.status].name

            if success is False:
                build_process_results.failure.append(
                    BulkRecordFail(
                        title=title,
                        error=status
                    )
                )
                logging.debug(f"Had problem adding record for {title}: {status}")
                continue

            rq = router_pb2.RouterRecordRotationRequest()
            rq.recordUid = url_safe_str_to_bytes(bulk_record.record_uid)
            rq.revision = 0

            # Set the gateway/configuration that this record should be connected.
            rq.configurationUid = url_safe_str_to_bytes(gateway_context.configuration_uid)

            # Only set the resource if the record type is a PAM User.
            # Machines, databases, and directories have a login/password in the record that indicates who the admin is.
            if bulk_record.record_type == "pamUser" and bulk_record.parent_record_uid is not None:
                rq.resourceUid = url_safe_str_to_bytes(bulk_record.parent_record_uid)

            # Right now, the schedule and password complexity are not set. This would be part of a rule engine.
            rq.schedule = ''
            rq.pwdComplexity = b''
            rq.disabled = rotation_disabled

            router_set_record_rotation_information(params, rq)

            created_cache.append(bulk_record.record_uid)

            build_process_results.success.append(
                BulkRecordSuccess(
                    title=title,
                    record_uid=bulk_record.record_uid
                )
            )
        print("")
        sys.stdout.flush()

        params.sync_data = True

        return build_process_results

    @classmethod
    def _convert_records(cls, bulk_convert_records: List[BulkRecordConvert], context: Optional[Any] = None):

        params = context.get("params")
        gateway_context = context.get("gateway_context")

        for bulk_convert_record in bulk_convert_records:

            record = vault.KeeperRecord.load(params, bulk_convert_record.record_uid)

            rotation_disabled = False

            rq = router_pb2.RouterRecordRotationRequest()
            rq.recordUid = url_safe_str_to_bytes(bulk_convert_record.record_uid)

            # We can't set the version to 0 if it's greater than 0, look up prior version.
            record_rotation_revision = params.record_rotation_cache.get(bulk_convert_record.record_uid)
            rq.revision = record_rotation_revision.get('revision') if record_rotation_revision else 0

            # Set the gateway/configuration that this record should be connected.
            rq.configurationUid = url_safe_str_to_bytes(gateway_context.configuration_uid)

            # Only set the resource if the record type is a PAM User.
            # Machines, databases, and directories have a login/password in the record that indicates who the admin is.
            if record.record_type == "pamUser" and bulk_convert_record.parent_record_uid is not None:
                rq.resourceUid = url_safe_str_to_bytes(bulk_convert_record.parent_record_uid)
            else:
                rq.resourceUid = None

            # Right now, the schedule and password complexity are not set. This would be part of a rule engine.
            rq.schedule = ''
            rq.pwdComplexity = b''
            rq.disabled = rotation_disabled

            router_set_record_rotation_information(params, rq)

        params.sync_data = True

    @staticmethod
    def _get_directory_info(domain: str,
                            skip_users: bool = False,
                            context: Optional[Any] = None) -> Optional[DirectoryInfo]:
        """
        Get information about this record from the vault records.

        """

        params = context.get("params")
        gateway_context = context.get("gateway_context")

        directory_info = DirectoryInfo()

        # Find the all directory records, in for this gateway, that have a domain that matches what we are looking for.
        for directory_record in vault_extensions.find_records(params, record_type="pamDirectory"):
            directory_record = vault.TypedRecord.load(params,
                                                      directory_record.record_uid)  # type: Optional[TypedRecord]

            info = params.record_rotation_cache.get(directory_record.record_uid)
            if info is None:
                continue

            # Make sure this user is part of this gateway.
            if info.get("configuration_uid") != gateway_context.configuration_uid:
                continue

            domain_field = directory_record.get_typed_field("text", label="domainName")
            if len(domain_field.value) == 0 or domain_field.value[0] == "":
                continue

            if domain_field.value[0].lower() != domain.lower():
                continue

            directory_info.directory_record_uids.append(directory_record.record_uid)

        if directory_info.has_directories is True and skip_users is False:

            for user_record in vault_extensions.find_records(params, record_type="pamUser"):
                info = params.record_rotation_cache.get(user_record.record_uid)
                if info is None:
                    continue

                if info.get("resource_uid") is None or info.get("resource_uid") == "":
                    continue

                # If the user's belongs to a directory, and add it to the directory user list.
                if info.get("resource_uid") in info.directory_record_uids:
                    directory_info.directory_user_record_uids.append(user_record.record_uid)

        return directory_info

    @staticmethod
    def remove_job(params: KeeperParams, configuration_record: KeeperRecord, job_id: str):

        try:
            jobs = Jobs(record=configuration_record, params=params)
            jobs.cancel(job_id)
            print(f"{bcolors.OKGREEN}No items left to process. Removing completed discovery job.{bcolors.ENDC}")
        except Exception as err:
            logging.error(err)
            print(f"{bcolors.FAIL}No items left to process. Failed to remove discovery job.{bcolors.ENDC}")

    def execute(self, params: KeeperParams, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        job_id = kwargs.get("job_id")
        add_all = kwargs.get("add_all", False)
        smart_add = kwargs.get("smart_add", False)

        # Right now, keep dry_run False. We might add it back in.
        dry_run = kwargs.get("dry_run", False)
        debug_level = kwargs.get("debug_level", 0)

        all_gateways = GatewayContext.all_gateways(params)

        configuration_records = list(vault_extensions.find_records(params, "pam.*Configuration"))
        for configuration_record in configuration_records:

            gateway_context = GatewayContext.from_configuration_uid(params=params,
                                                                    configuration_uid=configuration_record.record_uid,
                                                                    gateways=all_gateways)
            if gateway_context is None:
                continue

            record_cache = self._build_record_cache(
                params=params,
                gateway_context=gateway_context
            )

            # Get the current job.
            # There can only be one active job.
            # This will give us the sync point for the delta
            jobs = Jobs(record=configuration_record, params=params, logger=logging, debug_level=debug_level)
            job_item = jobs.current_job
            if job_item is None:
                continue

            # If this is not the job we are looking for, continue to the next gateway.
            if job_item.job_id != job_id:
                continue

            if job_item.end_ts is None:
                print(f'{bcolors.FAIL}Discovery job is currently running. Cannot process.{bcolors.ENDC}')
                return
            if job_item.success is False:
                print(f'{bcolors.FAIL}Discovery job failed. Cannot process.{bcolors.ENDC}')
                return

            process = Process(
                record=configuration_record,
                job_id=job_item.job_id,
                params=params,
                logger=logging,
                debug_level=debug_level,
            )

            if dry_run is True:
                if add_all is True:
                    logging.debug("dry run has been set, disable auto add.")
                    add_all = False

                print(f"{bcolors.HEADER}The DRY RUN flag has been set. The rule engine will not add any records. "
                      f"You will not be prompted to edit or add records.{bcolors.ENDC}")
                print("")

            if add_all is True:
                print(f"{bcolors.HEADER}The ADD ALL flag has been set. All found items will be added.{bcolors.ENDC}")
                print("")

            try:
                results = process.run(

                    # This method can get a record using the record UID
                    record_lookup_func=self._record_lookup,

                    # Prompt user the about adding records
                    prompt_func=self._prompt,

                    # Flag to auto add resources with credential, and all it users.
                    smart_add=smart_add,

                    # Prompt user for an admin for a resource
                    prompt_admin_func=self._prompt_admin,

                    # If quit, confirm if the user wants to add records
                    prompt_confirm_add_func=self._prompt_confirm_add,

                    # Prepare records and place in queue; does not add record to vault
                    record_prepare_func=self._prepare_record,

                    # Add record to the vault, protobuf, and record-linking graph
                    record_create_func=self._create_records,

                    # This function will take existing pamUser record and make them belong to this
                    #  gateway.
                    record_convert_func=self._convert_records,

                    # A function to get directory users
                    directory_info_func=self._get_directory_info,

                    # Pass method that will display auto added records.
                    auto_add_result_func=self._display_auto_add_results,

                    # Provides a cache of the record key to record UID.
                    record_cache=record_cache,

                    # Commander-specific context.
                    # Record link will be added by Process run as "record_link"
                    context={
                        "params": params,
                        "gateway_context": gateway_context,
                        "dry_run": dry_run,
                        "add_all": add_all
                    }
                )

                logging.debug(f"Results: {results}")

                print("")
                if results is not None and results.num_results > 0:
                    print(f"{bcolors.OKGREEN}Successfully added {results.success_count} "
                          f"record{'s' if results.success_count != 1 else ''}.{bcolors.ENDC}")
                    if results.has_failures is True:
                        print(f"{bcolors.FAIL}There were {results.failure_count} "
                              f"failure{'s' if results.failure_count != 1 else ''}.{bcolors.ENDC}")
                        for fail in results.failure:
                            print(f" * {fail.title}: {fail.error}")

                    if process.no_items_left is True:
                        self.remove_job(params=params, configuration_record=configuration_record, job_id=job_id)
                else:
                    print(f"{bcolors.FAIL}No records have been added.{bcolors.ENDC}")

            except NoDiscoveryDataException:
                print(f"{bcolors.OKGREEN}All items have been added for this discovery job.{bcolors.ENDC}")
                self.remove_job(params=params, configuration_record=configuration_record, job_id=job_id)

            except Exception as err:
                print(f"{bcolors.FAIL}Could not process discovery: {err}{bcolors.ENDC}")
                raise err

            return

        print(f"{bcolors.HEADER}Could not find the Discovery job.{bcolors.ENDC}")
        print("")



from __future__ import annotations
from enum import Enum
from pydantic import BaseModel
from typing import List, Optional, Union


class BaseEnum(Enum):

    @classmethod
    def find_enum(cls, value: Union[Enum, str, int], default: Optional[Enum] = None):
        if value is not None:
            for e in cls:
                if e == value or e.value == value:
                    return e
            if hasattr(cls, str(value).upper()) is True:
                return getattr(cls, value.upper())
        return default


class RefType(BaseEnum):
    # 0
    GENERAL = "general"
    # 1
    USER = "user"
    # 2
    DEVICE = "device"
    # 3
    REC = "rec"
    # 4
    FOLDER = "folder"
    # 5
    TEAM = "team"
    # 6
    ENTERPRISE = "enterprise"
    # 7
    PAM_DIRECTORY = "pam_directory"
    # 8
    PAM_MACHINE = "pam_machine"
    # 9
    PAM_DATABASE = "pam_database"
    # 10
    PAM_USER = "pam_user"
    # 11
    PAM_NETWORK = "pam_network"
    #12
    PAM_BROWSER = "pam_browser"

    def __str__(self):
        return self.value


class EdgeType(BaseEnum):

    """
    DAG data type enum

    * DATA - encrypted data
    * KEY - encrypted key
    * LINK - like a key, but not encrypted
    * ACL - unencrypted set of access control flags
    * DELETION - removal of the previous edge at the same coordinates
    * DENIAL - an element that was shared through graph relationship, can be explicitly denied
    * UNDENIAL - negates the effect of denial, bringing back the share

    """
    DATA = "data"
    KEY = "key"
    LINK = "link"
    ACL = "acl"
    DELETION = "deletion"
    DENIAL = "denial"
    UNDENIAL = "undenial"

    # To store discovery, you would need data and key. To store relationships between records after the discovery
    # data was converted, you use Link.

    def __str__(self) -> str:
        return str(self.value)


class SyncQuery(BaseModel):
    streamId: Optional[str] = None    # base64 of a user's ID who is syncing.
    deviceId: Optional[str] = None
    syncPoint: Optional[int] = None
    graphId: Optional[int] = 0


class SyncDataItem(BaseModel):
    ref: dict
    parentRef: Optional[dict] = None
    content: Optional[str] = None
    type: Optional[str] = None
    path: Optional[str] = None
    deletion: Optional[bool] = False


class SyncData(BaseModel):
    syncPoint: int
    data: List[SyncDataItem]
    hasMore: bool


class Ref(BaseModel):
    type: RefType
    value: str
    name: Optional[str] = None


# Translation for Key
class Key(BaseModel):
    id: Ref
    value: str


class DAGData(BaseModel):
    type: EdgeType
    ref: Ref
    parentRef: Optional[Ref] = None
    content: Optional[str] = None
    path: Optional[str] = None


class DataPayload(BaseModel):
    origin: Ref
    dataList: List
    graphId: Optional[int] = 0




#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from .auth import auth_check, policy_check
from .security import security_check, limiter
from ..decorators.logging import debug_decorator

__all__ = [
    'auth_check',
    'policy_check',
    'security_check',
    'debug_decorator',
    'limiter'
]


import abc
import sqlite3
from typing import Callable
from dataclasses import dataclass

from ..storage import types as storage_types, in_memory, sqlite, sqlite_dao


@dataclass
class PedmAdminSettings(storage_types.IUid[str]):
    key: str = ''
    value: str = ''
    def uid(self) -> str:
        return self.key


@dataclass
class PedmStorageDeployment(storage_types.IUid[str]):
    deployment_uid: str = ''
    encrypted_key: bytes = b''
    disabled: bool = False
    data: bytes = b''
    public_key: bytes = b''
    created: int = 0
    last_updated: int = 0
    def uid(self) -> str:
        return self.deployment_uid


@dataclass
class PedmStorageAgent(storage_types.IUid[str]):
    agent_uid: str = ''
    machine_id: str = ''
    deployment_uid: str = ''
    public_key: bytes = b''
    data: bytes = b''
    disabled: bool = False
    created: int = 0
    modified: int = 0
    def uid(self) -> str:
        return self.agent_uid


@dataclass
class PedmStoragePolicy(storage_types.IUid[str]):
    policy_uid: str = ''
    admin_data: bytes = b''
    data: bytes = b''
    key: bytes = b''
    disabled: bool = False
    created: int = 0
    updated: int = 0
    def uid(self) -> str:
        return self.policy_uid


@dataclass
class PedmStorageCollection(storage_types.IUid[str]):
    collection_uid: str = ''
    collection_type: int = 0
    data: bytes = b''
    created: int = 0
    def uid(self) -> str:
        return self.collection_uid


@dataclass
class PedmStorageCollectionLink(storage_types.IUidLink[str, str]):
    collection_uid: str = ''
    link_uid: str = ''
    link_type: int = 0
    def subject_uid(self) -> str:
        return self.collection_uid
    def object_uid(self) -> str:
        return self.link_uid


@dataclass
class PedmStorageApproval(storage_types.IUid[str]):
    approval_uid: str = ''
    approval_type: int = 0
    agent_uid: str = ''
    account_info: bytes = b''
    application_info: bytes = b''
    justification: bytes = b''
    expire_in: int = 0
    created: int = 0
    def uid(self) -> str:
        return self.approval_uid

@dataclass
class PedmStorageApprovalStatus(storage_types.IUid[str]):
    approval_uid: str = ''
    approval_status: int = 0
    enterprise_user_id: int = 0
    modified: int = 0
    def uid(self) -> str:
        return self.approval_uid


class IPedmStorage(abc.ABC):
    @property
    @abc.abstractmethod
    def settings(self) -> storage_types.IEntityStorage[PedmAdminSettings, str]:
        pass

    @property
    @abc.abstractmethod
    def deployments(self) -> storage_types.IEntityStorage[PedmStorageDeployment, str]:
        pass

    @property
    @abc.abstractmethod
    def agents(self) -> storage_types.IEntityStorage[PedmStorageAgent, str]:
        pass

    @property
    @abc.abstractmethod
    def policies(self) -> storage_types.IEntityStorage[PedmStoragePolicy, str]:
        pass

    @property
    @abc.abstractmethod
    def collections(self) -> storage_types.IEntityStorage[PedmStorageCollection, str]:
        pass

    @property
    @abc.abstractmethod
    def collection_links(self) -> storage_types.ILinkStorage[PedmStorageCollectionLink, str, str]:
        pass

    @property
    @abc.abstractmethod
    def approvals(self) -> storage_types.IEntityStorage[PedmStorageApproval, str]:
        pass

    @property
    @abc.abstractmethod
    def approval_status(self) -> storage_types.IEntityStorage[PedmStorageApprovalStatus, str]:
        pass

    @abc.abstractmethod
    def reset(self):
        pass


class MemoryPedmStorage(IPedmStorage):
    def __init__(self):
        self._settings = in_memory.InMemoryEntityStorage[PedmAdminSettings, str]()
        self._deployments = in_memory.InMemoryEntityStorage[PedmStorageDeployment, str]()
        self._agents = in_memory.InMemoryEntityStorage[PedmStorageAgent, str]()
        self._policies = in_memory.InMemoryEntityStorage[PedmStoragePolicy, str]()
        self._collections = in_memory.InMemoryEntityStorage[PedmStorageCollection, str]()
        self._collection_links = in_memory.InMemoryLinkStorage[PedmStorageCollectionLink, str, str]()
        self._approvals = in_memory.InMemoryEntityStorage[PedmStorageApproval, str]()
        self._approval_status = in_memory.InMemoryEntityStorage[PedmStorageApprovalStatus, str]()

    @property
    def settings(self) -> storage_types.IEntityStorage[PedmAdminSettings, str]:
        return self._settings

    @property
    def deployments(self) -> storage_types.IEntityStorage[PedmStorageDeployment, str]:
        return self._deployments

    @property
    def agents(self) -> storage_types.IEntityStorage[PedmStorageAgent, str]:
        return self._agents

    @property
    def policies(self) -> storage_types.IEntityStorage[PedmStoragePolicy, str]:
        return self._policies

    @property
    def collections(self) -> storage_types.IEntityStorage[PedmStorageCollection, str]:
        return self._collections

    @property
    def collection_links(self) -> storage_types.ILinkStorage[PedmStorageCollectionLink, str, str]:
        return self._collection_links

    @property
    def approvals(self) -> storage_types.IEntityStorage[PedmStorageApproval, str]:
        return self._approvals

    @property
    def approval_status(self) -> storage_types.IEntityStorage[PedmStorageApprovalStatus, str]:
        return self._approval_status

    def reset(self):
        self._settings.clear()
        self._deployments.clear()
        self._agents.clear()
        self._policies.clear()
        self._collections.clear()
        self._collection_links.clear()
        self._approvals.clear()
        self._approval_status.clear()


class SqlitePedmStorage(IPedmStorage):
    def __init__(self, get_connection: Callable[[], sqlite3.Connection], enterprise_id: int):
        self.get_connection = get_connection
        self.enterprise_id = enterprise_id
        self.owner_column = 'enterprise_id'
        setting_schema = sqlite_dao.TableSchema.load_schema(
            PedmAdminSettings, 'key', owner_column=self.owner_column, owner_type=int)
        deployment_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageDeployment, primary_key='deployment_uid', owner_column=self.owner_column, owner_type=int)
        agent_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageAgent, primary_key='agent_uid', owner_column=self.owner_column, owner_type=int)
        policy_schema = sqlite_dao.TableSchema.load_schema(
            PedmStoragePolicy, primary_key='policy_uid', owner_column=self.owner_column, owner_type=int)
        collection_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageCollection, primary_key='collection_uid', owner_column=self.owner_column, owner_type=int)
        collection_link_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageCollectionLink, primary_key=['collection_uid', 'link_uid'], indexes={'Link': 'link_uid'},
            owner_column=self.owner_column, owner_type=int)
        approval_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageApproval, primary_key='approval_uid', owner_column=self.owner_column, owner_type=int)
        approval_status_schema = sqlite_dao.TableSchema.load_schema(
            PedmStorageApprovalStatus, primary_key='approval_uid', owner_column=self.owner_column, owner_type=int)

        sqlite_dao.verify_database(
            self.get_connection(),(setting_schema, deployment_schema, agent_schema, policy_schema,
                                   collection_schema, collection_link_schema, approval_schema, approval_status_schema))

        self._settings = sqlite.SqliteEntityStorage(self.get_connection, setting_schema, owner=self.enterprise_id)
        self._deployments = sqlite.SqliteEntityStorage(self.get_connection, deployment_schema, owner=self.enterprise_id)
        self._agents = sqlite.SqliteEntityStorage(self.get_connection, agent_schema, owner=self.enterprise_id)
        self._policies = sqlite.SqliteEntityStorage(self.get_connection, policy_schema, owner=self.enterprise_id)
        self._collections = sqlite.SqliteEntityStorage(self.get_connection, collection_schema, owner=self.enterprise_id)
        self._collection_links = sqlite.SqliteLinkStorage(self.get_connection, collection_link_schema, owner=self.enterprise_id)
        self._approvals = sqlite.SqliteEntityStorage(self.get_connection, approval_schema, owner=self.enterprise_id)
        self._approval_status = sqlite.SqliteEntityStorage(self.get_connection, approval_status_schema, owner=self.enterprise_id)

    @property
    def settings(self) -> storage_types.IEntityStorage[PedmAdminSettings, str]:
        return self._settings

    @property
    def deployments(self) -> storage_types.IEntityStorage[PedmStorageDeployment, str]:
        return self._deployments

    @property
    def agents(self) -> storage_types.IEntityStorage[PedmStorageAgent, str]:
        return self._agents

    @property
    def policies(self) -> storage_types.IEntityStorage[PedmStoragePolicy, str]:
        return self._policies

    @property
    def collections(self) -> storage_types.IEntityStorage[PedmStorageCollection, str]:
        return self._collections

    @property
    def collection_links(self) -> storage_types.ILinkStorage[PedmStorageCollectionLink, str, str]:
        return self._collection_links

    @property
    def approvals(self) -> storage_types.IEntityStorage[PedmStorageApproval, str]:
        return self._approvals

    @property
    def approval_status(self) -> storage_types.IEntityStorage[PedmStorageApprovalStatus, str]:
        return self._approval_status

    def reset(self):
        self._settings.delete_all()
        self._deployments.delete_all()
        self._agents.delete_all()
        self._policies.delete_all()
        self._collections.delete_all()
        self._collection_links.delete_all()
        self._approvals.delete_all()
        self._approval_status.delete_all()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

def split_to_list(s: str, sep: str):
    """ Split a string to a list using the supplied delimiter/separator """
    return [e.strip() for e in s.split(sep) if e.strip()]



from __future__ import annotations
import argparse
from ..discover import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ...display import bcolors
from ... import vault, vault_extensions
from ...discovery_common.infrastructure import Infrastructure
from ...discovery_common.record_link import RecordLink
from ...discovery_common.user_service import UserService
from ...discovery_common.types import UserAcl, DiscoveryObject
from ...discovery_common.constants import PAM_USER, PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY
from ...keeper_dag import EdgeType
import time
import re
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMDebugInfoCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-debug-info')

    type_name_map = {
        PAM_USER: "PAM User",
        PAM_MACHINE: "PAM Machine",
        PAM_DATABASE: "PAM Database",
        PAM_DIRECTORY: "PAM Directory",
    }

    # The record to base everything on.
    parser.add_argument('--record-uid', '-i', required=True, dest='record_uid', action='store',
                        help='Keeper PAM record UID.')

    def get_parser(self):
        return PAMDebugInfoCommand.parser

    def execute(self, params: KeeperParams, **kwargs):

        record_uid = kwargs.get("record_uid")
        record = vault.KeeperRecord.load(params, record_uid)  # type: Optional[TypedRecord]
        if record is None:
            print(f"{bcolors.FAIL}Record does not exists.{bcolors.ENDC}")
            return

        if record.record_type not in ["pamUser", "pamMachine", "pamDatabase", "pamDirectory"]:
            if re.search(r'^pam.*Configuration$', record.record_type) is None:
                print(f"{bcolors.FAIL}The record is a {record.record_type}. This is not a PAM record.{bcolors.ENDC}")
                return

        resource_uid = None
        controller_uid = None

        record_rotation = params.record_rotation_cache.get(record_uid)

        # Rotation setting don't exist, check each configuration for an active record.
        if record_rotation is None:
            print(f"{bcolors.WARNING}PAM record does not have protobuf rotation settings, "
                  f"checking all configurations.{bcolors.ENDC}")

            configuration_records = list(vault_extensions.find_records(params, "pam.*Configuration"))
            if len(configuration_records) == 0:
                print(f"{bcolors.FAIL}Cannot find any PAM configuration records in the Vault{bcolors.ENDC}")

            for configuration_record in configuration_records:
                record_link = RecordLink(record=configuration_record, params=params)
                record_vertex = record_link.dag.get_vertex(record.record_uid)
                if record_vertex is not None and record_vertex.active is True:
                    controller_uid = configuration_record.record_uid
                    break
            if controller_uid is None:
                print(f"{bcolors.FAIL}Could not find the record in any record linking graph; "
                      f"checked all configuration records.{bcolors.ENDC}")
                return

        # Else just get information from the rotation settings
        else:

            controller_uid = record_rotation.get("configuration_uid")
            if controller_uid is None:
                print(f"{bcolors.FAIL}Record does not have the PAM Configuration set.{bcolors.ENDC}")
                return

            resource_uid = record_rotation.get("resource_uid")

        configuration_record = vault.KeeperRecord.load(params, controller_uid)  # type: Optional[TypedRecord]
        if configuration_record is None:
            print(f"{bcolors.FAIL}The configuration record {controller_uid} does not exist.{bcolors.ENDC}")
            return

        gateway_context = GatewayContext.from_configuration_uid(params, controller_uid)
        if gateway_context is None:
            print(f"{bcolors.FAIL}Could not find the gateway for configuration record.{controller_uid}{bcolors.ENDC}")
            return

        infra = Infrastructure(record=configuration_record, params=params)
        infra.load()
        record_link = RecordLink(record=configuration_record, params=params)
        user_service = UserService(record=configuration_record, params=params)

        print("")
        print(self._h("Record Information"))
        print(f"  {self._b('Record UID')}: {record_uid}")
        print(f"  {self._b('Record Title')}: {record.title}")
        print(f"  {self._b('Record Type')}: {record.record_type}")
        print(f"  {self._b('Configuration UID')}: {configuration_record.record_uid}")
        print(f"  {self._b('Configuration Key Bytes Hex')}: {configuration_record.record_key.hex()}")
        if resource_uid is not None:
            print(f"  {self._b('Resource UID')}: {resource_uid}")

        if gateway_context is not None:
            print(f"  {self._b('Gateway Name')}: {gateway_context.gateway_name}")
            print(f"  {self._b('Gateway UID')}: {gateway_context.gateway_uid}")
        else:
            print(f"  {self._f('Cannot get gateway information. Gateway may not be up.')}")
        print("")

        def _print_field(f):
            if f.type == "password":
                display_value = f"{bcolors.OKGREEN}Password is set{bcolors.ENDC}"
                if f.value == 0 or len(f.value) == 0:
                    display_value = f"{bcolors.WARNING}Password IS NOT set{bcolors.ENDC}"
                print(f"   * Type: {f.type}, Label: {f.label or bcolors.OKBLUE + 'NO LABEL' + bcolors.ENDC}, "
                      f"Value(s): {display_value}")
            elif f.label == "privatePEMKey":
                display_value = f"{bcolors.OKGREEN}Private Key is set{bcolors.ENDC}"
                if field.value == 0 or len(f.value) == 0:
                    display_value = f"{bcolors.WARNING}Private Key IS NOT set{bcolors.ENDC}"
                print(f"   * Type: {f.type}, Label: {f.label or bcolors.OKBLUE + 'NO LABEL' + bcolors.ENDC}, "
                      f"Value(s): {display_value}")
            elif f.type == "secret":
                display_value = f"{bcolors.OKGREEN}Secret value is set{bcolors.ENDC}"
                if field.value == 0 or len(f.value) == 0:
                    display_value = f"Secret value IS NOT set"
                print(f"   * Type: {f.type}, Label: {f.label or bcolors.OKBLUE + 'NO LABEL' + bcolors.ENDC}, "
                      f"Value(s): {display_value}")
            else:
                print(f"   * Type: {f.type}, Label: {f.label or bcolors.OKBLUE + 'NO LABEL' + bcolors.ENDC}, "
                      f"Value(s): {f.value}")

        print(self._h("Fields"))
        print(self._b("  Record Type Fields"))
        if record.fields is not None and len(record.fields) > 0:
            for field in record.fields:
                _print_field(field)
        else:
            print(f"{bcolors.FAIL}    Record does not have record type fields!{bcolors.ENDC}")
        print("")
        print(self._b("  Custom Fields"))
        if record.custom is not None and len(record.custom) > 0:
            for field in record.custom:
                _print_field(field)
        else:
            print("    Record does not have custom fields.")
        print("")

        discovery_vertices = infra.dag.search_content({"record_uid": record.record_uid})
        record_vertex = record_link.dag.get_vertex(record.record_uid)

        if record_vertex is not None:
            print(self._h("Record Linking"))
            record_parent_vertices = record_vertex.belongs_to_vertices()
            print(self._b("  Parent Records"))
            if len(record_parent_vertices) > 0:
                for record_parent_vertex in record_parent_vertices:

                    parent_record = vault.KeeperRecord.load(params,
                                                            record_parent_vertex.uid)  # type: Optional[TypedRecord]
                    if parent_record is None:
                        print(f"{bcolors.FAIL}   * Parent record {record_parent_vertex.uid} "
                              f"does not exists.{bcolors.ENDC}")
                        continue

                    acl_edge = record_vertex.get_edge(record_parent_vertex, EdgeType.ACL)
                    if acl_edge is not None:
                        acl_content = acl_edge.content_as_object(UserAcl)  # type: UserAcl
                        print(f"    * ACL to {self._n(parent_record.record_type)}; {parent_record.title}; "
                              f"{record_parent_vertex.uid}")
                        if acl_content.is_admin is True:
                            print(f"      . Is {self._gr('Admin')}")
                        if acl_content.belongs_to is True:
                            print(f"      . Belongs")
                        else:
                            print(f"      . Is {self._bl('Remote user')}")

                        if acl_content.rotation_settings is None:
                            print(f"{bcolors.FAIL}      . There are no rotation settings!{bcolors.ENDC}")
                        else:
                            if (acl_content.rotation_settings.schedule is None
                                    or acl_content.rotation_settings.schedule == ""):
                                print(f"      . No Schedule")
                            else:
                                print(f"      . Schedule = {acl_content.rotation_settings.get_schedule()}")

                            if (acl_content.rotation_settings.pwd_complexity is None
                                    or acl_content.rotation_settings.pwd_complexity == ""):
                                print(f"      . No Password Complexity")
                            else:
                                key_bytes = record.record_key
                                print(f"      . Password Complexity = "
                                      f"{acl_content.rotation_settings.get_pwd_complexity(key_bytes)}")
                            print(f"      . Disabled = {acl_content.rotation_settings.disabled}")
                            print(f"      . NOOP = {acl_content.rotation_settings.noop}")
                            print(f"      . SaaS Config Records = {acl_content.rotation_settings.saas_record_uid_list}")

                    elif record.record_type == PAM_USER:
                        print(f"{bcolors.FAIL}    * PAM User has NO acl!!!!!!{bcolors.ENDC}")

                    link_edge = record_vertex.get_edge(record_parent_vertex, EdgeType.LINK)
                    if link_edge is not None:
                        print(f"    * LINK to {self._n(parent_record.record_type)}; {parent_record.title}; "
                              f"{record_parent_vertex.uid}")
            else:
                # This really should not happen
                print(f"{bcolors.FAIL}   Record does not have a parent record.{bcolors.ENDC}")
            print("")

            record_child_vertices = record_vertex.has_vertices()
            print(self._b("  Child Records"))
            if len(record_child_vertices) > 0:
                for record_child_vertex in record_child_vertices:
                    child_record = vault.KeeperRecord.load(params,
                                                           record_child_vertex.uid)  # type: Optional[TypedRecord]

                    if child_record is None:
                        print(f"{bcolors.FAIL}    * Child record {record_child_vertex.uid} "
                              f"does not exists.{bcolors.ENDC}")
                        continue

                    acl_edge = record_child_vertex.get_edge(record_vertex, EdgeType.ACL)
                    link_edge = record_child_vertex.get_edge(record_vertex, EdgeType.LINK)
                    if acl_edge is not None:
                        acl_content = acl_edge.content_as_object(UserAcl)
                        print(f"    * ACL from {self._n(child_record.record_type)}; {child_record.title}; "
                              f"{record_child_vertex.uid}")
                        if acl_content.is_admin is True:
                            print(f"      . Is {self._gr('Admin')}")
                        if acl_content.belongs_to is True:
                            print(f"      . Belongs")
                        else:
                            print(f"      . Is {self._bl('Remote user')}")
                    elif link_edge is not None:
                        print(f"    * LINK from {self._n(child_record.record_type)}; {child_record.title}; "
                              "{record_child_vertex.uid}")
                    else:
                        for edge in record_vertex.edges:  # List[DAGEdge]
                            print(f"    * {self._f(edge.edge_type)}?")

            else:
                # This is OK
                print(f"    Record does not have any children.")
            print("")

        else:
            print(f"{bcolors.FAIL}Cannot find record in record linking.{bcolors.ENDC}")

        # Only PAM User and PAM Machine can have services and tasks.
        # This is really only Windows machines.
        if record.record_type == PAM_USER or record.record_type == PAM_MACHINE:

            # Get the user to service/task vertex.
            user_service_vertex = user_service.dag.get_vertex(record_uid)

            if user_service_vertex is not None:

                # If the record is a PAM User
                if record.record_type == PAM_USER:

                    user_results = {
                        "is_task": [],
                        "is_service": []
                    }

                    # Get a list of all the resources the user is the username/password on service/task.
                    for us_machine_vertex in user_service.get_resource_vertices(record_uid):

                        # Get the resource record
                        us_machine_record = (
                            vault.KeeperRecord.load(params, us_machine_vertex.uid))  # type: Optional[TypedRecord]

                        acl = user_service.get_acl(us_machine_vertex.uid, user_service_vertex.uid)
                        for attr in ["is_task", "is_service"]:
                            value = getattr(acl, attr)
                            if value is True:

                                # If the resource record does not exist.
                                if us_machine_record is None:

                                    # Default the title to Unknown (in red).
                                    # See if we have an infrastructure vertex with this record UID.
                                    # If we do have it, use the title inside the first vertex's data content.
                                    title = self._f("Unknown")
                                    infra_resource_vertices = infra.dag.search_content(
                                        {"record_uid": us_machine_vertex.uid})
                                    if len(infra_resource_vertices) > 0:
                                        infra_resource_vertex = infra_resource_vertices[0]
                                        if infra_resource_vertex.has_data is True:
                                            content = DiscoveryObject.get_discovery_object(infra_resource_vertex)
                                            title = content.title

                                    user_results[attr].append(f"  * Record {us_machine_vertex.uid}, "
                                                              f"{title} does not exists.")

                                # Record exists; just use information from the record.
                                else:
                                    user_results[attr].append(f"  * {us_machine_record.title}, "
                                                              f"{us_machine_vertex.uid}")

                    print(f"{bcolors.HEADER}Service on Machines{bcolors.ENDC}")
                    if len(user_results["is_service"]) > 0:
                        for service in user_results["is_service"]:
                            print(service)
                    else:
                        print("  PAM User is not used for any services.")
                    print("")

                    print(f"{bcolors.HEADER}Scheduled Tasks on Machines{bcolors.ENDC}")
                    if len(user_results["is_task"]) > 0:
                        for task in user_results["is_task"]:
                            print(task)
                    else:
                        print("  PAM User is not used for any scheduled tasks.")
                    print("")

                # If the record is a PAM Machine
                else:
                    user_results = {
                        "is_task": [],
                        "is_service": []
                    }

                    # Get the users that are used for tasks/services on this machine.
                    for us_user_vertex in user_service.get_user_vertices(record_uid):

                        us_user_record = vault.KeeperRecord.load(params,
                                                                 us_user_vertex.uid)  # type: Optional[TypedRecord]
                        acl = user_service.get_acl(user_service_vertex.uid, us_user_vertex.uid)
                        for attr in ["is_task", "is_service"]:
                            value = getattr(acl, attr)
                            if value is True:

                                # If the user record does not exist.
                                if us_user_record is None:

                                    # Default the title to Unknown (in red).
                                    # See if we have an infrastructure vertex with this record UID.
                                    # If we do have it, use the title inside the first vertex's data content.
                                    title = self._f("Unknown")
                                    infra_resource_vertices = infra.dag.search_content(
                                        {"record_uid": us_user_vertex.uid})
                                    if len(infra_resource_vertices) > 0:
                                        infra_resource_vertex = infra_resource_vertices[0]
                                        if infra_resource_vertex.has_data is True:
                                            content = DiscoveryObject.get_discovery_object(infra_resource_vertex)
                                            title = content.title

                                    user_results[attr].append(f"  * Record {us_user_vertex.uid}, "
                                                              f"{title} does not exists.")

                                # Record exists; just use information from the record.
                                else:
                                    user_results[attr].append(f"  * {us_user_record.title}, "
                                                              f"{us_user_vertex.uid}")

                    print(f"{bcolors.HEADER}Users that are used for Services{bcolors.ENDC}")
                    if len(user_results["is_service"]) > 0:
                        for service in user_results["is_service"]:
                            print(service)
                    else:
                        print("  Machine does not use any non-builtin users for services.")
                    print("")

                    print(f"{bcolors.HEADER}Users that are used for Scheduled Tasks{bcolors.ENDC}")
                    if len(user_results["is_task"]) > 0:
                        for task in user_results["is_task"]:
                            print(task)
                    else:
                        print("  Machine does not use any non-builtin users for scheduled tasks.")
                    print("")
            else:
                print(self._f("There are no services or schedule tasks associated with this record."))
                print("")
        try:
            if len(discovery_vertices) == 0:
                print(f"{bcolors.FAIL}Could not find any discovery infrastructure vertices for "
                      f"{record.record_uid}{bcolors.ENDC}")
            elif len(discovery_vertices) > 0:

                if len(discovery_vertices) > 1:
                    print(f"{bcolors.FAIL}Found multiple vertices with the record UID of "
                          f"{record.record_uid}{bcolors.ENDC}")
                    for vertex in discovery_vertices:
                        print(f" * Infrastructure Vertex UID: {vertex.uid}")
                    print("")

                discovery_vertex = discovery_vertices[0]
                content = DiscoveryObject.get_discovery_object(discovery_vertex)

                missing_since = "NA"
                if content.missing_since_ts is not None:
                    missing_since = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(content.missing_since_ts))

                print(self._h("Discovery Object Information"))
                print(f"  {self._b('Vertex UID')}: {content.uid}")
                print(f"  {self._b('Object ID')}: {content.id}")
                print(f"  {self._b('Record UID')}: {content.record_uid}")
                print(f"  {self._b('Parent Record UID')}: {content.parent_record_uid}")
                print(f"  {self._b('Shared Folder UID')}: {content.shared_folder_uid}")
                print(f"  {self._b('Record Type')}: {content.record_type}")
                print(f"  {self._b('Object Type')}: {content.object_type_value}")
                print(f"  {self._b('Ignore Object')}: {content.ignore_object}")
                print(f"  {self._b('Rule Engine Result')}: {content.action_rules_result}")
                print(f"  {self._b('Name')}: {content.name}")
                print(f"  {self._b('Generated Title')}: {content.title}")
                print(f"  {self._b('Generated Description')}: {content.description}")
                print(f"  {self._b('Missing Since')}: {missing_since}")
                print(f"  {self._b('Discovery Notes')}:")
                for note in content.notes:
                    print(f" * {note}")
                if content.error is not None:
                    print(f"{bcolors.FAIL}    Error: {content.error}{bcolors.ENDC}")
                    if content.stacktrace is not None:
                        print(f"{bcolors.FAIL}    Stack Trace:{bcolors.ENDC}")
                        print(f"{bcolors.FAIL}{content.stacktrace}{bcolors.ENDC}")
                print("")
                print(f"{bcolors.HEADER}Record Type Specifics{bcolors.ENDC}")

                if record.record_type == PAM_USER:
                    print(f"  {self._b('User')}: {content.item.user}")
                    print(f"  {self._b('DN')}: {content.item.dn}")
                    print(f"  {self._b('Database')}: {content.item.database}")
                    print(f"  {self._b('Active')}: {content.item.active}")
                    print(f"  {self._b('Expired')}: {content.item.expired}")
                    print(f"  {self._b('Source')}: {content.item.source}")
                elif record.record_type == PAM_MACHINE:
                    print(f"  {self._b('Host')}: {content.item.host}")
                    print(f"  {self._b('IP')}: {content.item.ip}")
                    print(f"  {self._b('Port')}: {content.item.port}")
                    print(f"  {self._b('Operating System')}: {content.item.os}")
                    print(f"  {self._b('Provider Region')}: {content.item.provider_region}")
                    print(f"  {self._b('Provider Group')}: {content.item.provider_group}")
                    print(f"  {self._b('Is the Gateway')}: {content.item.is_gateway}")
                    print(f"  {self._b('Allows Admin')}: {content.item.allows_admin}")
                    print(f"  {self._b('Admin Reason')}: {content.item.admin_reason}")
                    print("")
                    # If facts are not set, inside discover may not have been performed for the machine.
                    if content.item.facts.id is not None and content.item.facts.name is not None:
                        print(f"  {self._b('Machine Name')}: {content.item.facts.name}")
                        print(f"  {self._b('Machine ID')}: {content.item.facts.id.machine_id}")
                        print(f"  {self._b('Product ID')}: {content.item.facts.id.product_id}")
                        print(f"  {self._b('Board Serial')}: {content.item.facts.id.board_serial}")
                        print(f"  {self._b('Directories')}:")
                        if content.item.facts.directories is not None and len(content.item.facts.directories) > 0:
                            for directory in content.item.facts.directories:
                                print(f"    * Directory Domain: {directory.domain}")
                                print(f"      Software: {directory.software}")
                                print(f"      Login Format: {directory.login_format}")
                        else:
                            print("    Machines is not using any directories.")

                        print("")
                        print(f"  {self._b('Services')} (Non Builtin Users):")
                        if len(content.item.facts.services) > 0:
                            for service in content.item.facts.services:
                                print(f"    * {service.name} = {service.user}")
                        else:
                            print("    Machines has no services that are using non-builtin users.")

                        print(f"  {self._b('Scheduled Tasks')} (Non Builtin Users)")
                        if len(content.item.facts.tasks) > 0:
                            for task in content.item.facts.tasks:
                                print(f"    * {task.name} = {task.user}")
                        else:
                            print("    Machines has no schedules tasks that are using non-builtin users.")
                    else:
                        print(f"{bcolors.FAIL}    Machine facts are not set. Discover inside may not have been "
                              f"performed.{bcolors.ENDC}")
                elif record.record_type == PAM_DATABASE:
                    print(f"  {self._b('Host')}: {content.item.host}")
                    print(f"  {self._b('IP')}: {content.item.ip}")
                    print(f"  {self._b('Port')}: {content.item.port}")
                    print(f"  {self._b('Database Type')}: {content.item.type}")
                    print(f"  {self._b('Database')}: {content.item.database}")
                    print(f"  {self._b('Use SSL')}: {content.item.use_ssl}")
                    print(f"  {self._b('Provider Region')}: {content.item.provider_region}")
                    print(f"  {self._b('Provider Group')}: {content.item.provider_group}")
                    print(f"  {self._b('Allows Admin')}: {content.item.allows_admin}")
                    print(f"  {self._b('Admin Reason')}: {content.item.admin_reason}")
                elif record.record_type == PAM_DIRECTORY:
                    print(f"  {self._b('Host')}: {content.item.host}")
                    print(f"  {self._b('IP')}: {content.item.ip}")
                    print(f"  {self._b('Port')}: {content.item.port}")
                    print(f"  {self._b('Directory Type')}: {content.item.type}")
                    print(f"  {self._b('Use SSL')}: {content.item.use_ssl}")
                    print(f"  {self._b('Provider Region')}: {content.item.provider_region}")
                    print(f"  {self._b('Provider Group')}: {content.item.provider_group}")
                    print(f"  {self._b('Allows Admin')}: {content.item.allows_admin}")
                    print(f"  {self._b('Admin Reason')}: {content.item.admin_reason}")

                print("")
                print(self._h("Belongs To Vertices (Parents)"))
                vertices = discovery_vertex.belongs_to_vertices()
                for vertex in vertices:
                    content = DiscoveryObject.get_discovery_object(vertex)
                    print(f"  * {content.description} ({vertex.uid})")
                    for edge_type in [EdgeType.LINK, EdgeType.ACL, EdgeType.KEY, EdgeType.DELETION]:
                        edge = discovery_vertex.get_edge(vertex, edge_type=edge_type)
                        if edge is not None:
                            print(f"    . {edge_type}, active: {edge.active}")

                if len(vertices) == 0:
                    print(f"{bcolors.FAIL}  Does not belong to anyone{bcolors.ENDC}")

                print("")
                print(f"{bcolors.HEADER}Vertices Belonging To (Children){bcolors.ENDC}")
                vertices = discovery_vertex.has_vertices()
                for vertex in vertices:
                    content = DiscoveryObject.get_discovery_object(vertex)
                    print(f"  * {content.description} ({vertex.uid})")
                    for edge_type in [EdgeType.LINK, EdgeType.ACL, EdgeType.KEY, EdgeType.DELETION]:
                        edge = vertex.get_edge(discovery_vertex, edge_type=edge_type)
                        if edge is not None:
                            print(f"    . {edge_type}, active: {edge.active}")
                if len(vertices) == 0:
                    print(f"  Does not have any children.")

                print("")
            else:
                print(f"{bcolors.FAIL}Could not find infrastructure vertex.{bcolors.ENDC}")
        except Exception as err:
            print(f"{bcolors.FAIL}Could not get information on infrastructure: {err}{bcolors.ENDC}")



#!/usr/bin/env python3

"""Test keepercommander.subfolder."""

from unittest.mock import Mock

import pytest

import keepercommander.subfolder as subfolder


def BFN(*, type, uid, parent_uid, name, subfolders):
    """Build a mock BaseFolderNode."""
    result = Mock(name=name)
    result.type = type
    result.uid = uid
    result.parent_uid = parent_uid
    result.name = name
    # A list of UID's, not folders.
    assert all(isinstance(subfolder, str) for subfolder in subfolders)
    result.subfolders = subfolders
    return result


def folder_cache():
    """Build a two-node folder_cache.  Return it and the root uid."""
    cd_tests_uid = 'b' * 22
    cd_tests_bfn = BFN(type='user_folder', uid=cd_tests_uid, parent_uid=None, name='cd-tests', subfolders=[])
    root_uid = 'a' * 22
    root_bfn = BFN(type='/', uid=None, parent_uid=None, name='root', subfolders=[cd_tests_uid])
    dict_ = {
        root_uid: root_bfn,
        cd_tests_uid: cd_tests_bfn,
    }
    return dict_, root_bfn, cd_tests_bfn


def create_fake_params():
    """Create a fake params instance for testing."""
    params = Mock()
    (params.folder_cache, root_bfn, cd_tests_bfn) = folder_cache()
    params.current_folder = ''
    params.root_folder = root_bfn

    return params, root_bfn, cd_tests_bfn


(global_params, global_root_bfn, global_cd_tests_bfn) = create_fake_params()

global_test_params = (
    ('a', global_root_bfn, 'a'),
    ('/a', global_root_bfn, 'a'),
    ('/a/b', global_root_bfn, 'a/b'),
    ('/cd-tests/a', global_cd_tests_bfn, 'a'),
    ('/cd-tests', global_cd_tests_bfn, ''),
    ('a//b', global_root_bfn, 'a//b'),
    ('//a', global_root_bfn, '//a'),
    ('//a//b', global_root_bfn, '//a//b'),
    ('/cd-tests/a//b//c', global_cd_tests_bfn, 'a//b//c'),
    ('/cd-tests/..', global_root_bfn, ''),
    ('/cd-tests/.', global_cd_tests_bfn, ''),
    ('/..', global_root_bfn, ''),
    ('/.', global_root_bfn, ''),
    ('/./cd-tests', global_cd_tests_bfn, ''),
    ('/./cd-tests/nonexistent', global_cd_tests_bfn, 'nonexistent'),
    ('/./cd-tests/./nonexistent', global_cd_tests_bfn, 'nonexistent'),
    # The next three were complicating tab completion, so they no longer work.
    # ('/./cd-tests/ ', global_cd_tests_bfn, ''),
    # ('/ cd-tests', global_cd_tests_bfn, ''),
    # ('/ cd-tests ', global_cd_tests_bfn, ''),

    # This is a corner case we are willing to ignore
    # ('/ /a', global_root_bfn, '//a'),
    ('/', global_root_bfn, ''),
    ('//', global_root_bfn, '//'),
)


@pytest.mark.parametrize('input_, expected_folder, expected_final', global_test_params)
def test_subfolder_try_resolve_path(input_, expected_folder, expected_final):
    """Test try_resolve_path."""
    actual_folder, actual_final = subfolder.try_resolve_path(global_params, input_)
    assert actual_folder is expected_folder
    assert actual_final == expected_final
